// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// gren.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_gren_h
#define included_gren_h

////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
#define ISR inline static real
  //////////////////////////////////////////////////////////////////////////////
  // F_n kernels and their 1st derivative            r < eps                    
  //////////////////////////////////////////////////////////////////////////////
  struct F0 {
    ISR gd0(const real x) { return 1.5-0.5*x; }
    ISR gd1(const real x) { return -one; }
  };
  //----------------------------------------------------------------------------
  struct F1 {
    ISR gd0(const real x) { return (0.375*x-1.25)*x+1.875; }
    ISR gd1(const real x) { return 1.5*x-2.5; }
  };
  //----------------------------------------------------------------------------
  struct F2 {
    ISR gd0(const real x) { return ((-0.3125*x+1.3125)*x-2.1875)*x+2.1875; }
    ISR gd1(const real x) { return (-1.875*x+5.25)*x-4.375; }
  };
  //----------------------------------------------------------------------------
  struct F3 {
    ISR gd0(const real x) { return (((0.2734375*x-1.40625)*x+2.953125)*x
				    -3.28125)*x+2.4609375; }
    ISR gd1(const real x) { return ((2.1875*x-8.4375)*x+11.8125)*x-6.5625; }
  };
#if NDIM==3
  //////////////////////////////////////////////////////////////////////////////
  // K_n kernels and their 1st derivative in 3D      r < eps                    
  //////////////////////////////////////////////////////////////////////////////
  struct K0 {
    ISR gd0(const real x) { return (1.3125*x-3.125)*x+2.8125; }
    ISR gd1(const real x) { return 5.25*x-6.25; }
  };
  //----------------------------------------------------------------------------
  struct K1 {
    ISR gd0(const real x) { return ((-1.40625*x+4.59375)*x-5.46875)*x
			           +3.28125; }
    ISR gd1(const real x) { return (-8.4375*x+18.375)*x-10.9375; }
  };
  //----------------------------------------------------------------------------
  struct K2 {
    ISR gd0(const real x) { return (((1.50390625*x-6.328125)*x
				     +10.3359375)*x-8.203125)*x+3.69140625; }
    ISR gd1(const real x) { return ((12.03125*x-37.96875)*x+41.34375)*x
			            -16.40625; }
  };
  //----------------------------------------------------------------------------
  struct K3 {
    ISR gd0(const real x) { return ((((-1.599609375*x+8.271484375)*x
				      -17.40234375)*x+18.94921875)*x
				    -11.279296875)*x+4.060546875; }
    ISR gd1(const real x) { return (((-15.99609375*x+66.171875)*x
				     -104.4140625)*x+75.796875)*x
			            -22.55859375; }
  };
#else // NDIM==3
  //////////////////////////////////////////////////////////////////////////////
  // K_n kernels and their 1st derivative in 2D      r < eps                    
  //////////////////////////////////////////////////////////////////////////////
  struct K0 {
    ISR gd0(const real x) { return (2.25*x-5)*x+3.75; }
    ISR gd1(const real x) { return 9*x-10; }
  };
  //----------------------------------------------------------------------------
  struct K1 {
    ISR gd0(const real x) { return ((-2.5*x+7.875)*x-8.75)*x+4.375; }
    ISR gd1(const real x) { return (-15*x+31.5)*x-17.5; }
  };
  //----------------------------------------------------------------------------
  struct K2 {
    ISR gd0(const real x) { return (((2.734375*x-11.25)*x+17.71875)*x
				    -13.125)*x+4.921875; }
    ISR gd1(const real x) { return ((21.875*x-67.5)*x+70.875)*x-26.25; }
  };
  //----------------------------------------------------------------------------
  struct K3 {
    ISR gd0(const real x) { return ((((-2.953125*x+15.0390625)*x-30.9375)*x
				     +32.484375)*x-18.046875)*x+5.4140625; }
    ISR gd1(const real x) { return (((-29.53125*x+120.3125)*x-185.625)*x
				    +129.9375)*x-36.09375; }
  };
#endif // NDIM==3
#undef ISR
}
////////////////////////////////////////////////////////////////////////////////
// F_n or K_n kernel: compute pot & acc for any r, using fixes eps              
////////////////////////////////////////////////////////////////////////////////
#define HARDSOFT_GLOB(MUM,KERNEL)					\
  if(Rq<EPQ) {								\
    x  = Rq*IEQ;							\
    D0 = KERNEL::gd0(x) * (ME=MUM*IEP);					\
    R *= KERNEL::gd1(x) * (ME*IEQ);					\
  } else {								\
    x  = one/Rq;							\
    D0 = MUM*SQRT(x);							\
    R *=-D0*x;								\
  }
////////////////////////////////////////////////////////////////////////////////
// F_n or K_n kernel: compute pot & acc for any r, using individual eps_i       
////////////////////////////////////////////////////////////////////////////////
#define HARDSOFT_INDI(MUM,KERNEL)					\
  EPS= eps_mutual(A,B);							\
  EPQ= EPS*EPS;								\
  if(Rq<EPQ) {								\
    IEP= one/EPS;							\
    IEQ= IEP*IEP;							\
    x  = Rq*IEQ;							\
    D0 = KERNEL::gd0(x) * (ME=MUM*IEP);					\
    R *= KERNEL::gd1(x) * (ME*IEQ);					\
  } else {								\
    x  = one/Rq;							\
    D0 = MUM*SQRT(x);							\
    R *=-D0*x;								\
  }
////////////////////////////////////////////////////////////////////////////////
// Newtonian and P_n kernels: compute pot & acc for any r                       
////////////////////////////////////////////////////////////////////////////////
#define NEWTON(MUM)							\
  x  = one/Rq;								\
  D0 = MUM*SQRT(x);							\
  R *=-D0*x;
//------------------------------------------------------------------------------
#define P0(MUM)								\
  x  = one/(Rq+EPQ);							\
  D0 = MUM*SQRT(x);							\
  R *=-D0*x;
//------------------------------------------------------------------------------
#define P1(MUM)								\
  x  = one/(Rq+EPQ);							\
  D0 = MUM*SQRT(x);							\
  register real D1 =-D0*x;						\
  register real heq= half*EPQ;						\
  D0-= heq*D1;								\
  D1+= heq*3*D1*x;							\
  R *= D1;
//------------------------------------------------------------------------------
#define P2(MUM)								\
  x  = one/(Rq+EPQ);							\
  D0 = MUM*SQRT(x);							\
  register real D1 =-D0*x, D2=-3*D1*x, D3=-5*D2*x;			\
  register real heq= half*EPQ;						\
  D0-= heq*(D1-heq*D2);							\
  D1-= heq*(D2-heq*D3);							\
  R *= D1;
//------------------------------------------------------------------------------
#define P3(MUM)								\
  x  = one/(Rq+EPQ);							\
  D0 = MUM*SQRT(x);							\
  register real D1 =-D0*x, D2=-3*D1*x, D3=-5*D2*x, D4=-7*D3*x;		\
  register real heq= half*EPQ, hheq=half*heq;				\
  D0 -= ((heq*D3-D2)*hheq+D1)*heq;					\
  D1 -= ((heq*D4-D3)*hheq+D2)*heq;					\
  R  *= D1;
////////////////////////////////////////////////////////////////////////////////
// Newtonian and P_n kernels: allow for individual eps_i                        
////////////////////////////////////////////////////////////////////////////////
#define P_INDI(MUM,P_GLOB)						\
  EPS= eps_mutual(A,B);							\
  EPQ= EPS*EPS;								\
  P_GLOB(MUM)
////////////////////////////////////////////////////////////////////////////////
// body-body interactions with eps fixed                                        
////////////////////////////////////////////////////////////////////////////////
#define BODYBODY_GLOB(MUM)						\
switch(KERN) {								\
 case nbdy::f0: { HARDSOFT_GLOB(MUM,F0) } break;			\
 case nbdy::f1: { HARDSOFT_GLOB(MUM,F1) } break;			\
 case nbdy::f2: { HARDSOFT_GLOB(MUM,F2) } break;			\
 case nbdy::f3: { HARDSOFT_GLOB(MUM,F3) } break;			\
 case nbdy::k0: { HARDSOFT_GLOB(MUM,K0) } break;			\
 case nbdy::k1: { HARDSOFT_GLOB(MUM,K1) } break;			\
 case nbdy::k2: { HARDSOFT_GLOB(MUM,K2) } break;			\
 case nbdy::k3: { HARDSOFT_GLOB(MUM,K3) } break;			\
 case nbdy::p0: { P0(MUM)               } break;			\
 case nbdy::p1: { P1(MUM)               } break;			\
 case nbdy::p2: { P2(MUM)               } break;			\
 case nbdy::p3: { P3(MUM)               } break;			\
 case nbdy::newton:							\
 default:           { NEWTON(MUM)           } break;			\
}
////////////////////////////////////////////////////////////////////////////////
// body-body interactions with eps to be set using eps_mutual                   
////////////////////////////////////////////////////////////////////////////////
#define BODYBODY_INDI(MUM)						\
switch(KERN) {								\
 case nbdy::f0: { HARDSOFT_INDI(MUM,F0) } break;			\
 case nbdy::f1: { HARDSOFT_INDI(MUM,F1) } break;			\
 case nbdy::f2: { HARDSOFT_INDI(MUM,F2) } break;			\
 case nbdy::f3: { HARDSOFT_INDI(MUM,F3) } break;			\
 case nbdy::k0: { HARDSOFT_INDI(MUM,K0) } break;			\
 case nbdy::k1: { HARDSOFT_INDI(MUM,K1) } break;			\
 case nbdy::k2: { HARDSOFT_INDI(MUM,K2) } break;			\
 case nbdy::k3: { HARDSOFT_INDI(MUM,K3) } break;			\
 case nbdy::p0: { P_INDI(MUM,P0)        } break;			\
 case nbdy::p1: { P_INDI(MUM,P1)        } break;			\
 case nbdy::p2: { P_INDI(MUM,P2)        } break;			\
 case nbdy::p3: { P_INDI(MUM,P3)        } break;			\
 case nbdy::newton:							\
 default:           { NEWTON(MUM)           } break;			\
}
////////////////////////////////////////////////////////////////////////////////
// cell-node interactions with eps known                                        
////////////////////////////////////////////////////////////////////////////////
#define GREENS_FUNCTION							\
  register real	x,D0,D1,D2,D3;						\
  switch(KERN) {							\
  case nbdy::p3: {							\
    register real D4,D5,D6,heq=half*EPQ,hheq=half*heq;			\
    x  = one/(Rq+EPQ);							\
    D0 = mutual_mass(A,B)*SQRT(x);					\
    D1 =   -D0*x;							\
    D2 = -3*D1*x;							\
    D3 = -5*D2*x;							\
    D4 = -7*D3*x,							\
    D5 = -9*D4*x;							\
    D6 =-11*D5*x;							\
    D0-= ((heq*D3-D2)*hheq+D1)*heq;					\
    D1-= ((heq*D4-D3)*hheq+D2)*heq;					\
    D2-= ((heq*D5-D4)*hheq+D3)*heq;					\
    D3-= ((heq*D6-D5)*hheq+D4)*heq;					\
    } break;								\
  case nbdy::p2: {							\
    register real D4,D5,heq=half*EPQ;					\
    x  = one/(Rq+EPQ);							\
    D0 = mutual_mass(A,B)*SQRT(x);					\
    D1 =  -D0*x;							\
    D2 =-3*D1*x;							\
    D3 =-5*D2*x;							\
    D4 =-7*D3*x,							\
    D5 =-9*D4*x;							\
    D0-= heq*(D1-heq*D2);						\
    D1-= heq*(D2-heq*D3);						\
    D2-= heq*(D3-heq*D4);						\
    D3-= heq*(D4-heq*D5);						\
    } break;								\
  case nbdy::p1: {							\
    x  = one/(Rq+EPQ);							\
    D0 = mutual_mass(A,B)*SQRT(x);					\
    D1 =  -D0*x;							\
    D2 =-3*D1*x;							\
    D3 =-5*D2*x;							\
    register real heq=half*EPQ;						\
    D0 -= heq*D1;							\
    D1 -= heq*D2;							\
    D2 -= heq*D3;							\
    D3 += heq*7*D3*x;							\
    } break;								\
  case nbdy::p0:							\
    x  = one/(Rq+EPQ);							\
    D0 = mutual_mass(A,B)*SQRT(x);					\
    D1 =  -D0*x;							\
    D2 =-3*D1*x;							\
    D3 =-5*D2*x;							\
    break;								\
  default:								\
    x  = one/Rq;							\
    D0 = mutual_mass(A,B)*SQRT(x);					\
    D1 =  -D0*x;							\
    D2 =-3*D1*x;							\
    D3 =-5*D2*x;							\
  }
////////////////////////////////////////////////////////////////////////////////
#endif // included_gren_h
