// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// tens.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 1999-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_tens_h
#define included_tens_h

#ifndef included_vect_h
#include <nbdy/vect.h>
#endif

#ifndef NDIM
#error " NDIM not #defined in tens.h "
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
#if NDIM == 2

#define NMAT 3
#define CI(op,b,ox) A[0] op b[0] ox A[1] op b[1] ox A[2] op b[2]
#define CX(op,b,ox) A[0] op b    ox A[1] op b    ox A[2] op b
#define FI(op,b)    A[0] op b[0];   A[1] op b[1];   A[2] op b[2]
#define FX(op,b)    A[0] op b;      A[1] op b;      A[2] op b

#else

#define NMAT  6
#define CI(op,b,ox) A[0] op b[0] ox A[1] op b[1] ox A[2] op b[2] ox	\
                    A[3] op b[3] ox A[4] op b[4] ox A[5] op b[5]
#define CX(op,b,ox) A[0] op b    ox A[1] op b    ox A[2] op b    ox	\
                    A[3] op b    ox A[4] op b    ox A[5] op b
#define FI(op,b)    A[0] op b[0];   A[1] op b[1];   A[2] op b[2];	\
                    A[3] op b[3];   A[4] op b[4];   A[5] op b[5]
#define FX(op,b)    A[0] op b;      A[1] op b;      A[2] op b;		\
                    A[3] op b;      A[4] op b;      A[5] op b

#endif

#define OP      operator
  //////////////////////////////////////////////////////////////////////////////
  // some forward declarations                                                  
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL> class sym3;
  template<typename REAL> class esym2;
  template<typename REAL> class esym3;
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::sym2                                                           
  //----------------------------------------------------------------------------
  // a symmetric 3x3 matrix                                                     
  // element mapping is as follows:   0 -> 0,0                                  
  //                                  1 -> 0,1 & 1,0                            
  //                                  2 -> 0,2 & 2,0                            
  //                                  3 -> 1,1                                  
  //                                  4 -> 1,2 & 2,1                            
  //                                  5 -> 2,2 & 2,2                            
  //----------------------------------------------------------------------------
  // or a symmetric 2x2 matrix                                                  
  // element mapping is as follows:   0 -> 0,0                                  
  //                                  1 -> 0,1 & 1,0                            
  //                                  2 -> 1,1                                  
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL>                      // scalar type                   
  class sym2 {
    //--------------------------------------------------------------------------
    // friendships                                                              
    //--------------------------------------------------------------------------
    friend class sym3<REAL>;
    friend class esym2<REAL>;
    friend class esym3<REAL>;
    //--------------------------------------------------------------------------
    // types                                                                    
    //--------------------------------------------------------------------------
  protected:
    typedef REAL                    real;      // type for const returns        
    typedef vector<REAL>            vect;      // type for const returns        
    typedef sym2                    matr;      // type for const args           
    typedef const int             c_indx;      // type for const args           
    typedef const REAL           &c_real;      // type for const args           
    typedef const vector<REAL>   &c_vect;      // type for const args           
    typedef const evector<REAL>  &c_evec;      // type for const args           
    typedef const sym2           &c_matr;      // type for const args           
    typedef const esym2<REAL>    &c_emat;      // type for const args           
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
  private:
    REAL A[NMAT];
    //--------------------------------------------------------------------------
    // construction                                                             
    //--------------------------------------------------------------------------
  public:
    sym2 ()         {}
    sym2 (c_real x) { FX(=,x); }
    sym2 (c_matr x) { FI(=,x.A); }
    sym2 (c_emat x) { FI(=,x.A); }
    //--------------------------------------------------------------------------
    // non-const methods (operator=() to be defined in derived class)           
    //--------------------------------------------------------------------------
    real* pointer()                   { return A; }
    real &OP[]   (c_indx i)           { return A[i]; }
    matr &OP=    (c_real x)           { FX(=,x); return *this; }
    matr &OP=    (c_matr x)           { FI(=,x.A); return *this; }
    matr &OP=    (c_emat x)           { FI(=,x.A); return *this; }
    matr &OP+=   (c_matr x)           { FI(+=,x.A); return *this; }
    matr &OP+=   (c_emat x)           { FI(+=,x.A); return *this; }
    matr &OP-=   (c_matr x)           { FI(-=,x.A); return *this; }
    matr &OP-=   (c_emat x)           { FI(-=,x.A); return *this; }
    matr &OP*=   (c_real x)           { FX(*=,x); return *this; }
    matr &OP/=   (c_real x)           { return OP*=(real(1)/x); }
    matr &negate ()                   { FI(=-,A); return *this; }
    matr &add_mul(c_matr y, c_real x) { FI(+=,x*y.A); return *this; }
    matr &sub_mul(c_matr y, c_real x) { FI(-=,x*y.A); return *this; }
    matr &add_mul(c_emat y, c_real x) { FI(+=,x*y.A); return *this; }
    matr &sub_mul(c_emat y, c_real x) { FI(-=,x*y.A); return *this; }
    //--------------------------------------------------------------------------
#if NDIM==2
    matr &unity ()         { A[0]=A[2]=1; A[1]=0; return *this; }
    matr &unity (c_real x) { A[0]=A[2]=x; A[1]=0; return *this; }
    matr &add_unity(c_real x) { A[0]+=x; A[2]+=x; return *this; }
#else
    matr &unity ()         { A[0]=A[3]=A[5]=1; A[1]=A[2]=A[4]=0; return *this; }
    matr &unity (c_real x) { A[0]=A[3]=A[5]=x; A[1]=A[2]=A[4]=0; return *this; }
    matr &add_unity(c_real x) { A[0]+=x; A[3]+=x; A[5]+=x; return *this; }
#endif
    //--------------------------------------------------------------------------
    matr &apply  (real(*f)(c_real))   { 
      A[0]=f(A[0]); A[1]=f(A[1]); A[2]=f(A[2]);
#if NDIM==3
      A[3]=f(A[3]); A[4]=f(A[4]); A[5]=f(A[5]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    matr &connect(c_matr x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]); A[1]=f(A[1],x.A[1]); A[2]=f(A[2],x.A[2]);
#if NDIM==3
      A[3]=f(A[3],x.A[3]); A[4]=f(A[4],x.A[4]); A[5]=f(A[5],x.A[5]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    matr &connect(c_emat x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]); A[1]=f(A[1],x.A[1]); A[2]=f(A[2],x.A[2]);
#if NDIM==3
      A[3]=f(A[3],x.A[3]); A[4]=f(A[4],x.A[4]); A[5]=f(A[5],x.A[5]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    matr&outprod(c_vect x) {
      // outer product Aij = Xi*Xj                                              
      A[0]=x.A[0]*x.A[0];
      A[1]=x.A[0]*x.A[1];
#if NDIM==2
      A[2]=x.A[1]*x.A[1];
#else
      A[2]=x.A[0]*x.A[2]; 
      A[3]=x.A[1]*x.A[1];
      A[4]=x.A[1]*x.A[2];
      A[5]=x.A[2]*x.A[2];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    matr&outprod(c_evec x) {
      // outer product Aij = Xi*Xj                                              
      A[0]=x.A[0]*x.A[0];
      A[1]=x.A[0]*x.A[1];
#if NDIM==2
      A[2]=x.A[1]*x.A[1];
#else
      A[2]=x.A[0]*x.A[2]; 
      A[3]=x.A[1]*x.A[1];
      A[4]=x.A[1]*x.A[2];
      A[5]=x.A[2]*x.A[2];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    matr&outprod(c_vect x, c_vect y) {
      // outer product Aij = Xi*Yj + Xj*Yi                                      
      A[0]=twice(x.A[0]*y.A[0]);
      A[1]=x.A[0]*y.A[1] + x.A[1]*y.A[0];
#if NDIM==2
      A[2]=twice(x.A[1]*y.A[1]);
#else
      A[2]=x.A[0]*y.A[2] + x.A[2]*y.A[0];
      A[3]=twice(x.A[1]*y.A[1]);
      A[4]=x.A[1]*y.A[2] + x.A[2]*y.A[1];
      A[5]=twice(x.A[2]*y.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    matr&outprod(c_vect x, c_evec y) {
      // outer product Aij = Xi*Yj + Xj*Yi                                      
      A[0]=twice(x.A[0]*y.A[0]);
      A[1]=x.A[0]*y.A[1] + x.A[1]*y.A[0];
#if NDIM==2
      A[2]=twice(x.A[1]*y.A[1]);
#else
      A[2]=x.A[0]*y.A[2] + x.A[2]*y.A[0];
      A[3]=twice(x.A[1]*y.A[1]);
      A[4]=x.A[1]*y.A[2] + x.A[2]*y.A[1];
      A[5]=twice(x.A[2]*y.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    matr&outprod(c_evec x, c_vect y) {
      // outer product Aij = Xi*Yj + Xj*Yi                                      
      A[0]=twice(x.A[0]*y.A[0]);
      A[1]=x.A[0]*y.A[1] + x.A[1]*y.A[0];
#if NDIM==2
      A[2]=twice(x.A[1]*y.A[1]);
#else
      A[2]=x.A[0]*y.A[2] + x.A[2]*y.A[0];
      A[3]=twice(x.A[1]*y.A[1]);
      A[4]=x.A[1]*y.A[2] + x.A[2]*y.A[1];
      A[5]=twice(x.A[2]*y.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    matr&outprod(c_evec x, c_evec y) {
      // outer product Aij = Xi*Yj + Xj*Yi                                      
      A[0]=twice(x.A[0]*y.A[0]);
      A[1]=x.A[0]*y.A[1] + x.A[1]*y.A[0];
#if NDIM==2
      A[2]=twice(x.A[1]*y.A[1]);
#else
      A[2]=x.A[0]*y.A[2] + x.A[2]*y.A[0];
      A[3]=twice(x.A[1]*y.A[1]);
      A[4]=x.A[1]*y.A[2] + x.A[2]*y.A[1];
      A[5]=twice(x.A[2]*y.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    matr&quadrup(c_vect x, c_real a, c_real b) {
      // Aij = a*D_ij + b*Xi*Xj                                                 
      register real tmp=b*x.A[0];
      A[0]= a+tmp*x.A[0];
      A[1]=   tmp*x.A[1];
#if NDIM==2
      A[2]= a+b*x.A[1]*x.A[1];
#else
      A[2]=   tmp*x.A[2];
      tmp =     b*x.A[1];
      A[3]= a+tmp*x.A[1];
      A[4]=   tmp*x.A[2];
      A[5]= a+b*x.A[2]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    matr&quadrup(c_evec x, c_real a, c_real b) {
      // Aij = a*D_ij + b*Xi*Xj                                                 
      register real tmp=b*x.A[0];
      A[0]= a+tmp*x.A[0];
      A[1]=   tmp*x.A[1];
#if NDIM==2
      A[2]= a+b*x.A[1]*x.A[1];
#else
      A[2]=   tmp*x.A[2];
      tmp =     b*x.A[1];
      A[3]= a+tmp*x.A[1];
      A[4]=   tmp*x.A[2];
      A[5]= a+b*x.A[2]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    matr&add_quadrup(c_vect x, c_real a, c_real b) {
      // Aij += a*D_ij + b*Xi*Xj                                                
      register real tmp=b*x.A[0];
      A[0]+= a+tmp*x.A[0];
      A[1]+=   tmp*x.A[1];
#if NDIM==2
      A[2]+= a+b*x.A[1]*x.A[1];
#else
      A[2]+=   tmp*x.A[2];
      tmp  =     b*x.A[1];
      A[3]+= a+tmp*x.A[1];
      A[4]+=   tmp*x.A[2];
      A[5]+= a+b*x.A[2]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    matr&add_quadrup(c_evec x, c_real a, c_real b) {
      // Aij += a*D_ij + b*Xi*Xj                                                
      register real tmp=b*x.A[0];
      A[0]+= a+tmp*x.A[0];
      A[1]+=   tmp*x.A[1];
#if NDIM==2
      A[2]+= a+b*x.A[1]*x.A[1];
#else
      A[2]+=   tmp*x.A[2];
      tmp  =     b*x.A[1];
      A[3]+= a+tmp*x.A[1];
      A[4]+=   tmp*x.A[2];
      A[5]+= a+b*x.A[2]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    // const methods                                                            
    //--------------------------------------------------------------------------
    const real* const_pointer()     const  { return A; }
    real const&OP[](c_indx i) const { return A[i]; }
    matr OP*       (c_real x) const { register matr y(*this); y*=x; return y; }
    matr OP/       (c_real x) const { register matr y(*this); y*=1/x; return y; }
    matr OP+       (c_matr x) const { register matr y(*this); y+=x; return y; }
    matr OP+       (c_emat x) const { register matr y(*this); y+=x; return y; }
    matr OP-       (c_matr x) const { register matr y(*this); y-=x; return y; }
    matr OP-       (c_emat x) const { register matr y(*this); y-=x; return y; }
    matr OP-       ()         const { register matr y(0); y-=*this; return y; }
    bool OP==      (c_real x) const { return CX(==,x,&&); }
    bool OP==      (c_matr x) const { return CI(==,x.A,&&); }
    bool OP==      (c_emat x) const { return CI(==,x.A,&&); }
    bool OP!=      (c_real x) const { return CX(!=,x,||); }
    bool OP!=      (c_matr x) const { return CI(!=,x.A,||); }
    bool OP!=      (c_emat x) const { return CI(!=,x.A,||); }
    //--------------------------------------------------------------------------
    real maxnorm() const {
      return max(abs(A[0]), max(abs(A[1]),
#if NDIM==2
				abs(A[2])
#else
				max(abs(A[2]), max(abs(A[3]),
						   max(abs(A[4]),abs(A[5]))))
#endif
				));
    }
    //--------------------------------------------------------------------------
    real trace() const {
      // trace = Aii                                                            
#if NDIM==2
      return A[0]+A[2];
#else
      return A[0]+A[3]+A[5];
#endif
    }
    //--------------------------------------------------------------------------
    real contract(c_vect x) const {
      // double inner product with vector:  result_i = Xj*Xi*Aij                
#if NDIM==2
      return x.A[0]*(A[0]*x.A[0]+twice(A[1]*x.A[1]))
	    +x.A[1]* A[2]*x.A[1];
#else
      return x.A[0]*(x.A[0]*A[0]+twice(x.A[1]*A[1]+x.A[2]*A[2]))
	    +x.A[1]*(x.A[1]*A[3]+twice(x.A[2]*A[4]))
	    +x.A[2]* x.A[2]*A[5];
#endif
    }  
    //--------------------------------------------------------------------------
    real contract(c_evec x) const {
      // double inner product with vector:  result_i = Xj*Xi*Aij                
#if NDIM==2
      return x.A[0]*(A[0]*x.A[0]+twice(A[1]*x.A[1]))
	    +x.A[1]* A[2]*x.A[1];
#else
      return x.A[0]*(x.A[0]*A[0]+twice(x.A[1]*A[1]+x.A[2]*A[2]))
	    +x.A[1]*(x.A[1]*A[3]+twice(x.A[2]*A[4]))
	    +x.A[2]* x.A[2]*A[5];
#endif
    }  
    //--------------------------------------------------------------------------
    vect OP* (c_vect x) const { 
      // inner product with vector:  result_i = Xj*Aij                          
      register vect z;
#if NDIM==2
      z.A[0] = A[0]*x.A[0] + A[1]*x.A[1];
      z.A[1] = A[1]*x.A[0] + A[2]*x.A[1];
#else
      z.A[0] = A[0]*x.A[0] + A[1]*x.A[1] + A[2]*x.A[2];
      z.A[1] = A[1]*x.A[0] + A[3]*x.A[1] + A[4]*x.A[2];
      z.A[2] = A[2]*x.A[0] + A[4]*x.A[1] + A[5]*x.A[2];
#endif
      return z;
    }
    //--------------------------------------------------------------------------
    vect OP* (c_evec x) const { 
      // inner product with vector:  result_i = Xj*Aij                          
      register vect z;
#if NDIM==2
      z.A[0] = A[0]*x.A[0] + A[1]*x.A[1];
      z.A[1] = A[1]*x.A[0] + A[2]*x.A[1];
#else
      z.A[0] = A[0]*x.A[0] + A[1]*x.A[1] + A[2]*x.A[2];
      z.A[1] = A[1]*x.A[0] + A[3]*x.A[1] + A[4]*x.A[2];
      z.A[2] = A[2]*x.A[0] + A[4]*x.A[1] + A[5]*x.A[2];
#endif
      return z;
    }
    //--------------------------------------------------------------------------
    real OP* (c_matr x) const {
      // inner product with matrix:  result_i = Aij*Xij                         
#if NDIM==2
      return twice(A[1]*x.A[1])+A[1]*x.A[1]; 
#else
      return  twice(A[1]*x.A[1] + A[2]*x.A[2] + A[4]*x.A[4])
                  + A[0]*x.A[0] + A[3]*x.A[3] + A[5]*x.A[5];
#endif
    }
    //--------------------------------------------------------------------------
    real OP* (c_emat x) const {
      // inner product with matrix:  result_i = Aij*Xij                         
#if NDIM==2
      return twice(A[1]*x.A[1])+A[1]*x.A[1]; 
#else
      return  twice(A[1]*x.A[1] + A[2]*x.A[2] + A[4]*x.A[4])
                  + A[0]*x.A[0] + A[3]*x.A[3] + A[5]*x.A[5];
#endif
    }
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::esym2                                                          
  //----------------------------------------------------------------------------
  // a symmetric 3x3 matrix                                                     
  // element mapping is as follows:   0 -> 0,0                                  
  //                                  1 -> 0,1 & 1,0                            
  //                                  2 -> 0,2 & 2,0                            
  //                                  3 -> 1,1                                  
  //                                  4 -> 1,2 & 2,1                            
  //                                  5 -> 2,2 & 2,2                            
  //----------------------------------------------------------------------------
  // or a symmetric 2x2 matrix                                                  
  // element mapping is as follows:   0 -> 0,0                                  
  //                                  1 -> 0,1 & 1,0                            
  //                                  2 -> 1,1                                  
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL>                      // scalar type                   
  class esym2 {
    //--------------------------------------------------------------------------
    // friendships                                                              
    //--------------------------------------------------------------------------
    friend class sym2<REAL>;
    friend class sym3<REAL>;
    friend class esym3<REAL>;
    //--------------------------------------------------------------------------
    // types                                                                    
    //--------------------------------------------------------------------------
  protected:
    typedef REAL                    real;      // type for const returns        
    typedef vector<REAL>            vect;      // type for const returns        
    typedef sym2<REAL>              matr;      // type for const args           
    typedef esym2                   emat;      // type for non-const return     
    typedef const int             c_indx;      // type for const args           
    typedef const REAL           &c_real;      // type for const args           
    typedef const vector<REAL>   &c_vect;      // type for const args           
    typedef const evector<REAL>  &c_evec;      // type for const args           
    typedef const sym2<REAL>     &c_matr;      // type for const args           
    typedef const esym2          &c_emat;      // type for const args           
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
  private:
    REAL *A;
    //--------------------------------------------------------------------------
    // construction                                                             
    //--------------------------------------------------------------------------
  public:
    esym2 (real* x) : A(x) {}
    //--------------------------------------------------------------------------
    // non-const methods (operator=() to be defined in derived class)           
    //--------------------------------------------------------------------------
    real* pointer()                   { return A; }
    real &OP[]   (c_indx i)           { return A[i]; }
    emat &OP=    (c_real x)           { FX(=,x); return *this; }
    emat &OP=    (c_matr x)           { FI(=,x.A); return *this; }
    emat &OP=    (c_emat x)           { FI(=,x.A); return *this; }
    emat &OP+=   (c_matr x)           { FI(+=,x.A); return *this; }
    emat &OP+=   (c_emat x)           { FI(+=,x.A); return *this; }
    emat &OP-=   (c_matr x)           { FI(-=,x.A); return *this; }
    emat &OP-=   (c_emat x)           { FI(-=,x.A); return *this; }
    emat &OP*=   (c_real x)           { FX(*=,x); return *this; }
    emat &OP/=   (c_real x)           { return OP*=(real(1)/x); }
    emat &negate ()                   { FI(=-,A); return *this; }
    emat &add_mul(c_matr y, c_real x) { FI(+=,x*y.A); return *this; }
    emat &sub_mul(c_matr y, c_real x) { FI(-=,x*y.A); return *this; }
    emat &add_mul(c_emat y, c_real x) { FI(+=,x*y.A); return *this; }
    emat &sub_mul(c_emat y, c_real x) { FI(-=,x*y.A); return *this; }
    //--------------------------------------------------------------------------
#if NDIM==2
    emat &unity ()         { A[0]=A[2]=1; A[1]=0; return *this; }
    emat &unity (c_real x) { A[0]=A[2]=x; A[1]=0; return *this; }
    emat &add_unity(c_real x) { A[0]+=x; A[2]+=x; return *this; }
#else
    emat &unity ()         { A[0]=A[3]=A[5]=1; A[1]=A[2]=A[4]=0; return *this; }
    emat &unity (c_real x) { A[0]=A[3]=A[5]=x; A[1]=A[2]=A[4]=0; return *this; }
    emat &add_unity(c_real x) { A[0]+=x; A[3]+=x; A[5]+=x; return *this; }
#endif
    //--------------------------------------------------------------------------
    emat &apply  (real(*f)(c_real))   { 
      A[0]=f(A[0]); A[1]=f(A[1]); A[2]=f(A[2]);
#if NDIM==3
      A[3]=f(A[3]); A[4]=f(A[4]); A[5]=f(A[5]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    emat &connect(c_matr x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]); A[1]=f(A[1],x.A[1]); A[2]=f(A[2],x.A[2]);
#if NDIM==3
      A[3]=f(A[3],x.A[3]); A[4]=f(A[4],x.A[4]); A[5]=f(A[5],x.A[5]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    emat &connect(c_emat x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]); A[1]=f(A[1],x.A[1]); A[2]=f(A[2],x.A[2]);
#if NDIM==3
      A[3]=f(A[3],x.A[3]); A[4]=f(A[4],x.A[4]); A[5]=f(A[5],x.A[5]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    emat &outprod(c_vect x) {
      // outer product Aij = Xi*Xj                                              
      A[0]=x.A[0]*x.A[0];
      A[1]=x.A[0]*x.A[1];
#if NDIM==2
      A[2]=x.A[1]*x.A[1];
#else
      A[2]=x.A[0]*x.A[2]; 
      A[3]=x.A[1]*x.A[1];
      A[4]=x.A[1]*x.A[2];
      A[5]=x.A[2]*x.A[2];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    emat &outprod(c_evec x) {
      // outer product Aij = Xi*Xj                                              
      A[0]=x.A[0]*x.A[0];
      A[1]=x.A[0]*x.A[1];
#if NDIM==2
      A[2]=x.A[1]*x.A[1];
#else
      A[2]=x.A[0]*x.A[2]; 
      A[3]=x.A[1]*x.A[1];
      A[4]=x.A[1]*x.A[2];
      A[5]=x.A[2]*x.A[2];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    emat &outprod(c_vect x, c_vect y) {
      // outer product Aij = Xi*Yj + Xj*Yi                                      
      A[0]=twice(x.A[0]*y.A[0]);
      A[1]=x.A[0]*y.A[1] + x.A[1]*y.A[0];
#if NDIM==2
      A[2]=twice(x.A[1]*y.A[1]);
#else
      A[2]=x.A[0]*y.A[2] + x.A[2]*y.A[0];
      A[3]=twice(x.A[1]*y.A[1]);
      A[4]=x.A[1]*y.A[2] + x.A[2]*y.A[1];
      A[5]=twice(x.A[2]*y.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    emat &outprod(c_vect x, c_evec y) {
      // outer product Aij = Xi*Yj + Xj*Yi                                      
      A[0]=twice(x.A[0]*y.A[0]);
      A[1]=x.A[0]*y.A[1] + x.A[1]*y.A[0];
#if NDIM==2
      A[2]=twice(x.A[1]*y.A[1]);
#else
      A[2]=x.A[0]*y.A[2] + x.A[2]*y.A[0];
      A[3]=twice(x.A[1]*y.A[1]);
      A[4]=x.A[1]*y.A[2] + x.A[2]*y.A[1];
      A[5]=twice(x.A[2]*y.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    emat &outprod(c_evec x, c_vect y) {
      // outer product Aij = Xi*Yj + Xj*Yi                                      
      A[0]=twice(x.A[0]*y.A[0]);
      A[1]=x.A[0]*y.A[1] + x.A[1]*y.A[0];
#if NDIM==2
      A[2]=twice(x.A[1]*y.A[1]);
#else
      A[2]=x.A[0]*y.A[2] + x.A[2]*y.A[0];
      A[3]=twice(x.A[1]*y.A[1]);
      A[4]=x.A[1]*y.A[2] + x.A[2]*y.A[1];
      A[5]=twice(x.A[2]*y.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    emat &outprod(c_evec x, c_evec y) {
      // outer product Aij = Xi*Yj + Xj*Yi                                      
      A[0]=twice(x.A[0]*y.A[0]);
      A[1]=x.A[0]*y.A[1] + x.A[1]*y.A[0];
#if NDIM==2
      A[2]=twice(x.A[1]*y.A[1]);
#else
      A[2]=x.A[0]*y.A[2] + x.A[2]*y.A[0];
      A[3]=twice(x.A[1]*y.A[1]);
      A[4]=x.A[1]*y.A[2] + x.A[2]*y.A[1];
      A[5]=twice(x.A[2]*y.A[2]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    emat &quadrup(c_vect x, c_real a, c_real b) {
      // Aij = a*D_ij + b*Xi*Xj                                                 
      register real tmp=b*x.A[0];
      A[0]= a+tmp*x.A[0];
      A[1]=   tmp*x.A[1];
#if NDIM==2
      A[2]= a+b*x.A[1]*x.A[1];
#else
      A[2]=   tmp*x.A[2];
      tmp =     b*x.A[1];
      A[3]= a+tmp*x.A[1];
      A[4]=   tmp*x.A[2];
      A[5]= a+b*x.A[2]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    emat &quadrup(c_evec x, c_real a, c_real b) {
      // Aij = a*D_ij + b*Xi*Xj                                                 
      register real tmp=b*x.A[0];
      A[0]= a+tmp*x.A[0];
      A[1]=   tmp*x.A[1];
#if NDIM==2
      A[2]= a+b*x.A[1]*x.A[1];
#else
      A[2]=   tmp*x.A[2];
      tmp =     b*x.A[1];
      A[3]= a+tmp*x.A[1];
      A[4]=   tmp*x.A[2];
      A[5]= a+b*x.A[2]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    emat &add_quadrup(c_vect x, c_real a, c_real b) {
      // Aij += a*D_ij + b*Xi*Xj                                                
      register real tmp=b*x.A[0];
      A[0]+= a+tmp*x.A[0];
      A[1]+=   tmp*x.A[1];
#if NDIM==2
      A[2]+= a+b*x.A[1]*x.A[1];
#else
      A[2]+=   tmp*x.A[2];
      tmp  =     b*x.A[1];
      A[3]+= a+tmp*x.A[1];
      A[4]+=   tmp*x.A[2];
      A[5]+= a+b*x.A[2]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    emat &add_quadrup(c_evec x, c_real a, c_real b) {
      // Aij += a*D_ij + b*Xi*Xj                                                
      register real tmp=b*x.A[0];
      A[0]+= a+tmp*x.A[0];
      A[1]+=   tmp*x.A[1];
#if NDIM==2
      A[2]+= a+b*x.A[1]*x.A[1];
#else
      A[2]+=   tmp*x.A[2];
      tmp  =     b*x.A[1];
      A[3]+= a+tmp*x.A[1];
      A[4]+=   tmp*x.A[2];
      A[5]+= a+b*x.A[2]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    // const methods                                                            
    //--------------------------------------------------------------------------
    const real* const_pointer()     const  { return A; }
    real const&OP[](c_indx i) const { return A[i]; }
    matr OP*       (c_real x) const { register matr y(*this); y*=x; return y; }
    matr OP/       (c_real x) const { register matr y(*this); y*=1/x; return y; }
    matr OP+       (c_matr x) const { register matr y(*this); y+=x; return y; }
    matr OP+       (c_emat x) const { register matr y(*this); y+=x; return y; }
    matr OP-       (c_matr x) const { register matr y(*this); y-=x; return y; }
    matr OP-       (c_emat x) const { register matr y(*this); y-=x; return y; }
    matr OP-       ()         const { register matr y(0); y-=*this; return y; }
    bool OP==      (c_real x) const { return CX(==,x,&&); }
    bool OP==      (c_matr x) const { return CI(==,x.A,&&); }
    bool OP==      (c_emat x) const { return CI(==,x.A,&&); }
    bool OP!=      (c_real x) const { return CX(!=,x,||); }
    bool OP!=      (c_matr x) const { return CI(!=,x.A,||); }
    bool OP!=      (c_emat x) const { return CI(!=,x.A,||); }
    //--------------------------------------------------------------------------
    real maxnorm() const {
      return max(abs(A[0]), max(abs(A[1]),
#if NDIM==2
				abs(A[2])
#else
				max(abs(A[2]), max(abs(A[3]),
						   max(abs(A[4]),abs(A[5]))))
#endif
				));
    }
    //--------------------------------------------------------------------------
    real trace() const {
      // trace = Aii                                                            
#if NDIM==2
      return A[0]+A[2];
#else
      return A[0]+A[3]+A[5];
#endif
    }
    //--------------------------------------------------------------------------
    real contract(c_vect x) const {
      // double inner product with vector:  result_i = Xj*Xi*Aij                
#if NDIM==2
      return x.A[0]*(A[0]*x.A[0]+twice(A[1]*x.A[1]))
	    +x.A[1]* A[2]*x.A[1];
#else
      return x.A[0]*(x.A[0]*A[0]+twice(x.A[1]*A[1]+x.A[2]*A[2]))
	    +x.A[1]*(x.A[1]*A[3]+twice(x.A[2]*A[4]))
	    +x.A[2]* x.A[2]*A[5];
#endif
    }  
    //--------------------------------------------------------------------------
    real contract(c_evec x) const {
      // double inner product with vector:  result_i = Xj*Xi*Aij                
#if NDIM==2
      return x.A[0]*(A[0]*x.A[0]+twice(A[1]*x.A[1]))
	    +x.A[1]* A[2]*x.A[1];
#else
      return x.A[0]*(x.A[0]*A[0]+twice(x.A[1]*A[1]+x.A[2]*A[2]))
	    +x.A[1]*(x.A[1]*A[3]+twice(x.A[2]*A[4]))
	    +x.A[2]* x.A[2]*A[5];
#endif
    }  
    //--------------------------------------------------------------------------
    vect OP* (c_vect x) const { 
      // inner product with vector:  result_i = Xj*Aij                          
      register vect z;
#if NDIM==2
      z.A[0] = A[0]*x.A[0] + A[1]*x.A[1];
      z.A[1] = A[1]*x.A[0] + A[2]*x.A[1];
#else
      z.A[0] = A[0]*x.A[0] + A[1]*x.A[1] + A[2]*x.A[2];
      z.A[1] = A[1]*x.A[0] + A[3]*x.A[1] + A[4]*x.A[2];
      z.A[2] = A[2]*x.A[0] + A[4]*x.A[1] + A[5]*x.A[2];
#endif
      return z;
    }
    //--------------------------------------------------------------------------
    vect OP* (c_evec x) const { 
      // inner product with vector:  result_i = Xj*Aij                          
      register vect z;
#if NDIM==2
      z.A[0] = A[0]*x.A[0] + A[1]*x.A[1];
      z.A[1] = A[1]*x.A[0] + A[2]*x.A[1];
#else
      z.A[0] = A[0]*x.A[0] + A[1]*x.A[1] + A[2]*x.A[2];
      z.A[1] = A[1]*x.A[0] + A[3]*x.A[1] + A[4]*x.A[2];
      z.A[2] = A[2]*x.A[0] + A[4]*x.A[1] + A[5]*x.A[2];
#endif
      return z;
    }
    //--------------------------------------------------------------------------
    real OP* (c_matr x) const {
      // inner product with matrix:  result_i = Aij*Xij                         
#if NDIM==2
      return twice(A[1]*x.A[1])+A[1]*x.A[1]; 
#else
      return  twice(A[1]*x.A[1] + A[2]*x.A[2] + A[4]*x.A[4])
                  + A[0]*x.A[0] + A[3]*x.A[3] + A[5]*x.A[5];
#endif
    }
    //--------------------------------------------------------------------------
    real OP* (c_emat x) const {
      // inner product with matrix:  result_i = Aij*Xij                         
#if NDIM==2
      return twice(A[1]*x.A[1])+A[1]*x.A[1]; 
#else
      return  twice(A[1]*x.A[1] + A[2]*x.A[2] + A[4]*x.A[4])
                  + A[0]*x.A[0] + A[3]*x.A[3] + A[5]*x.A[5];
#endif
    }
  };
  //////////////////////////////////////////////////////////////////////////////
  // related functions (coding as friends results in compiler problems)         
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL> inline
  vector<REAL> OP*(const vector<REAL>&x, const esym2<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  vector<REAL> OP*(const vector<REAL>&x, const sym2<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  vector<REAL> OP*(const evector<REAL>&x, const esym2<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  vector<REAL> OP*(const evector<REAL>&x, const sym2<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL trace(const sym2<REAL>&x) { return x.trace(); }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL trace(const esym2<REAL>&x) { return x.trace(); }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL maxnorm(const sym2<REAL>&x) { return x.maxnorm(); }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  REAL maxnorm(const esym2<REAL>&x) { return x.maxnorm(); }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  std::ostream& OP<<(std::ostream&s, const sym2<REAL>& x) {
    return util::write_array<REAL,NMAT>(s,x.const_pointer());
  }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  std::ostream& OP<<(std::ostream&s, const esym2<REAL>& x) {
    return util::write_array<REAL,NMAT>(s,x.const_pointer());
  }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  std::istream& OP>>(std::istream&s, sym2<REAL>& x) {
    return util::read_array<REAL,NMAT>(s,x.pointer());
  }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  std::istream& OP>>(std::istream&s, esym2<REAL>& x) {
    return util::read_array<REAL,NMAT>(s,x.pointer());
  }
  //////////////////////////////////////////////////////////////////////////////
#undef CI
#undef CX
#undef FI
#undef FX
#if NDIM == 2

#define NTEN 4
#define CI(op,b,ox) A[0] op b[0] ox A[1] op b[1] ox A[2] op b[2] ox A[3] op b[3]
#define CX(op,b,ox) A[0] op b    ox A[1] op b    ox A[2] op b    ox A[3] op b
#define FI(op,b)    A[0] op b[0];   A[1] op b[1];   A[2] op b[2];   A[3] op b[3]
#define FX(op,b)    A[0] op b;      A[1] op b;      A[2] op b;      A[3] op b

#else

#define NTEN 10
#define CI(op,b,ox) A[0] op b[0] ox A[1] op b[1] ox A[2] op b[2] ox A[3] op b[3] \
                 ox A[4] op b[4] ox A[5] op b[5] ox A[6] op b[6] ox A[7] op b[7] \
                 ox A[8] op b[8] ox A[9] op b[9]
#define CX(op,b,ox) A[0] op b    ox A[1] op b    ox A[2] op b    ox A[3] op b    \
                 ox A[4] op b    ox A[5] op b    ox A[6] op b    ox A[7] op b    \
                 ox A[8] op b    ox A[9] op b
#define FI(op,b) A[0] op b[0];   A[1] op b[1];   A[2] op b[2];   A[3] op b[3];	\
                 A[4] op b[4];   A[5] op b[5];   A[6] op b[6];   A[7] op b[7];	\
                 A[8] op b[8];   A[9] op b[9]
#define FX(op,b) A[0] op b;      A[1] op b;      A[2] op b;      A[3] op b;	\
                 A[4] op b;      A[5] op b;      A[6] op b;      A[7] op b;	\
                 A[8] op b;      A[9] op b

#endif
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::sym3                                                           
  //----------------------------------------------------------------------------
  // a symmetric 3x3x3 tensor                                                   
  // element mapping is as follows:   0 -> 0,0,0                                
  //                                  1 -> 0,0,1 & permutations                 
  //                                  2 -> 0,0,2 & permutations                 
  //                                  3 -> 0,1,1 & permutations                 
  //                                  4 -> 0,1,2 & permutations                 
  //                                  5 -> 0,2,2 & permutations                 
  //                                  6 -> 1,1,1                                
  //                                  7 -> 1,1,2 & permutations                 
  //                                  8 -> 1,2,2 & permutations                 
  //                                  9 -> 2,2,2                                
  //----------------------------------------------------------------------------
  // or a symmetric 2x2x2 tensor                                                
  // element mapping is as follows:   0 -> 0,0,0                                
  //                                  1 -> 0,0,1 & permutations                 
  //                                  2 -> 0,1,1 & permutations                 
  //                                  3 -> 1,1,1                                
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL>                      // scalar type                   
  class sym3 {
    //--------------------------------------------------------------------------
    // friendships                                                              
    //--------------------------------------------------------------------------
    friend class esym3<REAL>;
    //--------------------------------------------------------------------------
    // types                                                                    
    //--------------------------------------------------------------------------
  protected:
    typedef REAL                    real;      // type for const returns        
    typedef vector<REAL>            vect;      // type for const returns        
    typedef sym2<REAL>              matr;      // type for const args           
    typedef sym3                    tens;      // type for const args           
    typedef const int             c_indx;      // type for const args           
    typedef const REAL           &c_real;      // type for const args           
    typedef const vector<REAL>   &c_vect;      // type for const args           
    typedef const evector<REAL>  &c_evec;      // type for const args           
    typedef const sym2<REAL>     &c_matr;      // type for const args           
    typedef const esym2<REAL>    &c_emat;      // type for const args           
    typedef const sym3           &c_tens;      // type for const args           
    typedef const esym3<REAL>    &c_eten;      // type for const args           
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
  private:
    REAL A[NTEN];
    //--------------------------------------------------------------------------
    // construction                                                             
    //--------------------------------------------------------------------------
  public:
    sym3 ()         {}
    sym3 (c_real x) { FX(=,x); }
    sym3 (c_tens x) { FI(=,x.A); }
    sym3 (c_eten x) { FI(=,x.A); }
    //--------------------------------------------------------------------------
    // non-const methods                                                        
    //--------------------------------------------------------------------------
  public:
    real* pointer()                   { return A; }
    real &OP[]   (c_indx i)           { return A[i]; }
    tens &OP+=   (c_tens x)           { FI(+=,x.A); return *this; }
    tens &OP+=   (c_eten x)           { FI(+=,x.A); return *this; }
    tens &OP-=   (c_tens x)           { FI(-=,x.A); return *this; }
    tens &OP-=   (c_eten x)           { FI(-=,x.A); return *this; }
    tens &OP*=   (c_real x)           { FX(*=,x); return *this; }
    tens &OP/=   (c_real x)           { return OP*=(real(1)/x); }
    tens &negate ()                   { FI(=-,A); return *this; }
    tens &add_mul(c_tens y, c_real x) { FI(+=,x*y.A); return *this; }
    tens &add_mul(c_eten y, c_real x) { FI(+=,x*y.A); return *this; }
    tens &sub_mul(c_tens y, c_real x) { FI(-=,x*y.A); return *this; }
    tens &sub_mul(c_eten y, c_real x) { FI(-=,x*y.A); return *this; }
    //--------------------------------------------------------------------------
#if NDIM==2
    tens &unity  ()           { A[0]=A[3]=1; A[1]=A[2]=0; return *this; }
    tens &unity  (c_real x)   { A[0]=A[3]=x; A[1]=A[2]=0; return *this; }
    tens &add_unity(c_real x) { A[0]+=x; A[3]+=x;         return *this; }
#else
    //--------------------------------------------------------------------------
    tens &unity  () {
      A[1]=A[2]=A[3]=A[4]=A[5]=A[7]=A[8]=0;
      A[0]=A[6]=A[9]=1;
      return *this;
    }
    tens &unity  (c_real x) {
      A[1]=A[2]=A[3]=A[4]=A[5]=A[7]=A[8]=0;
      A[0]=A[6]=A[9]=x;
      return *this;
    }
    tens &add_unity  (c_real x) {
      A[0]+= x;
      A[6]+= x;
      A[9]+= x;
      return *this;
    }
#endif
    //--------------------------------------------------------------------------
    tens &apply  (real(*f)(c_real))   { 
      A[0]=f(A[0]); A[1]=f(A[1]); A[2]=f(A[2]); A[3]=f(A[3]); 
#if NDIM==3
      A[4]=f(A[4]); A[5]=f(A[5]); A[6]=f(A[6]); A[7]=f(A[7]);
      A[8]=f(A[8]); A[9]=f(A[9]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &connect(c_tens x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]);
      A[1]=f(A[1],x.A[1]);
      A[2]=f(A[2],x.A[2]);
      A[3]=f(A[3],x.A[3]); 
#if NDIM==3
      A[4]=f(A[4],x.A[4]);
      A[6]=f(A[5],x.A[6]);
      A[7]=f(A[7],x.A[7]);
      A[8]=f(A[8],x.A[8]);
      A[9]=f(A[9],x.A[9]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &connect(c_eten x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]);
      A[1]=f(A[1],x.A[1]);
      A[2]=f(A[2],x.A[2]);
      A[3]=f(A[3],x.A[3]); 
#if NDIM==3
      A[4]=f(A[4],x.A[4]);
      A[6]=f(A[5],x.A[6]);
      A[7]=f(A[7],x.A[7]);
      A[8]=f(A[8],x.A[8]);
      A[9]=f(A[9],x.A[9]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &outprod(c_vect x) {
      // outer product Aijk= Xi*Xj*Xk                                           
      register real tmp=x.A[0]*x.A[0];
      A[0] = tmp*x.A[0];
      A[1] = tmp*x.A[1];
#if NDIM==2
      tmp  = x.A[1]*x.A[1];
      A[2] = tmp*x.A[0];
      A[3] = tmp*x.A[1];
#else
      A[2] = tmp*x.A[2];
      tmp  = x.A[1]*x.A[1];
      A[3] = tmp*x.A[0];
      A[4] = x.A[0]*x.A[1]*x.A[2];
      A[5] = tmp*x.A[0];
      A[6] = tmp*x.A[1];
      tmp  = x.A[2]*x.A[2];
      A[7] = tmp*x.A[2];
      A[8] = tmp*x.A[1];
      A[9] = tmp*x.A[2];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &outprod(c_evec x) {
      // outer product Aijk= Xi*Xj*Xk                                           
      register real tmp=x.A[0]*x.A[0];
      A[0] = tmp*x.A[0];
      A[1] = tmp*x.A[1];
#if NDIM==2
      tmp  = x.A[1]*x.A[1];
      A[2] = tmp*x.A[0];
      A[3] = tmp*x.A[1];
#else
      A[2] = tmp*x.A[2];
      tmp  = x.A[1]*x.A[1];
      A[3] = tmp*x.A[0];
      A[4] = x.A[0]*x.A[1]*x.A[2];
      A[5] = tmp*x.A[0];
      A[6] = tmp*x.A[1];
      tmp  = x.A[2]*x.A[2];
      A[7] = tmp*x.A[2];
      A[8] = tmp*x.A[1];
      A[9] = tmp*x.A[2];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &outprod(c_vect x, c_matr y) {
      // outer product Aij = Xi*Yjk + Xj*Yik + Xk*Yij                           
      A[0] = trice(y.A[0]*x.A[0]);
      A[1] = twice(y.A[1]*x.A[0]) + y.A[0]*x.A[1];
#if NDIM==2
      A[2] = twice(y.A[1]*x.A[1]) + y.A[2]*x.A[0];
      A[3] = trice(y.A[2]*x.A[1]);
#else
      A[6] = trice(y.A[3]*x.A[1]);
      A[9] = trice(y.A[5]*x.A[2]);
      A[2] = twice(y.A[2]*x.A[0]) + y.A[0]*x.A[2];
      A[3] = twice(y.A[1]*x.A[1]) + y.A[3]*x.A[0];
      A[5] = twice(y.A[2]*x.A[2]) + y.A[5]*x.A[0];
      A[7] = twice(y.A[4]*x.A[1]) + y.A[3]*x.A[2];
      A[8] = twice(y.A[4]*x.A[2]) + y.A[5]*x.A[1];
      A[4] = y.A[1]*x.A[2] +y.A[2]*x.A[1] +y.A[4]*x.A[0];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &outprod(c_vect x, c_emat y) {
      // outer product Aij = Xi*Yjk + Xj*Yik + Xk*Yij                           
      A[0] = trice(y.A[0]*x.A[0]);
      A[1] = twice(y.A[1]*x.A[0]) + y.A[0]*x.A[1];
#if NDIM==2
      A[2] = twice(y.A[1]*x.A[1]) + y.A[2]*x.A[0];
      A[3] = trice(y.A[2]*x.A[1]);
#else
      A[6] = trice(y.A[3]*x.A[1]);
      A[9] = trice(y.A[5]*x.A[2]);
      A[2] = twice(y.A[2]*x.A[0]) + y.A[0]*x.A[2];
      A[3] = twice(y.A[1]*x.A[1]) + y.A[3]*x.A[0];
      A[5] = twice(y.A[2]*x.A[2]) + y.A[5]*x.A[0];
      A[7] = twice(y.A[4]*x.A[1]) + y.A[3]*x.A[2];
      A[8] = twice(y.A[4]*x.A[2]) + y.A[5]*x.A[1];
      A[4] = y.A[1]*x.A[2] +y.A[2]*x.A[1] +y.A[4]*x.A[0];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &outprod(c_evec x, c_matr y) {
      // outer product Aij = Xi*Yjk + Xj*Yik + Xk*Yij                           
      A[0] = trice(y.A[0]*x.A[0]);
      A[1] = twice(y.A[1]*x.A[0]) + y.A[0]*x.A[1];
#if NDIM==2
      A[2] = twice(y.A[1]*x.A[1]) + y.A[2]*x.A[0];
      A[3] = trice(y.A[2]*x.A[1]);
#else
      A[6] = trice(y.A[3]*x.A[1]);
      A[9] = trice(y.A[5]*x.A[2]);
      A[2] = twice(y.A[2]*x.A[0]) + y.A[0]*x.A[2];
      A[3] = twice(y.A[1]*x.A[1]) + y.A[3]*x.A[0];
      A[5] = twice(y.A[2]*x.A[2]) + y.A[5]*x.A[0];
      A[7] = twice(y.A[4]*x.A[1]) + y.A[3]*x.A[2];
      A[8] = twice(y.A[4]*x.A[2]) + y.A[5]*x.A[1];
      A[4] = y.A[1]*x.A[2] +y.A[2]*x.A[1] +y.A[4]*x.A[0];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &outprod(c_evec x, c_emat y) {
      // outer product Aij = Xi*Yjk + Xj*Yik + Xk*Yij                           
      A[0] = trice(y.A[0]*x.A[0]);
      A[1] = twice(y.A[1]*x.A[0]) + y.A[0]*x.A[1];
#if NDIM==2
      A[2] = twice(y.A[1]*x.A[1]) + y.A[2]*x.A[0];
      A[3] = trice(y.A[2]*x.A[1]);
#else
      A[6] = trice(y.A[3]*x.A[1]);
      A[9] = trice(y.A[5]*x.A[2]);
      A[2] = twice(y.A[2]*x.A[0]) + y.A[0]*x.A[2];
      A[3] = twice(y.A[1]*x.A[1]) + y.A[3]*x.A[0];
      A[5] = twice(y.A[2]*x.A[2]) + y.A[5]*x.A[0];
      A[7] = twice(y.A[4]*x.A[1]) + y.A[3]*x.A[2];
      A[8] = twice(y.A[4]*x.A[2]) + y.A[5]*x.A[1];
      A[4] = y.A[1]*x.A[2] +y.A[2]*x.A[1] +y.A[4]*x.A[0];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &quadrup(c_vect x, c_real a) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi)                                  
      register real ai = a*x.A[0];
      A[0] = ai+ai+ai;
#if NDIM==2
      A[2] = ai;
      ai   = a*x.A[1];
      A[3] = ai+ai+ai;
      A[1] = ai;
#else
      A[3] = ai;
      A[5] = ai;
      ai   = a*x.A[1];
      A[6] = ai+ai+ai;
      A[1] = ai;
      A[8] = ai;
      ai   = a*x.A[2];
      A[9] = ai+ai+ai;
      A[2] = ai;
      A[4] = 0;
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &quadrup(c_evec x, c_real a) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi)                                  
      register real ai = a*x.A[0];
      A[0] = ai+ai+ai;
#if NDIM==2
      A[2] = ai;
      ai   = a*x.A[1];
      A[3] = ai+ai+ai;
      A[1] = ai;
#else
      A[3] = ai;
      A[5] = ai;
      ai   = a*x.A[1];
      A[6] = ai+ai+ai;
      A[1] = ai;
      A[8] = ai;
      ai   = a*x.A[2];
      A[9] = ai+ai+ai;
      A[2] = ai;
      A[4] = 0;
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &add_quadrup(c_vect x, c_real a) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi)                                  
      register real ai = a*x.A[0];
      A[0]+= ai+ai+ai;
#if NDIM==2
      A[2]+= ai;
      ai   = a*x.A[1];
      A[3]+= ai+ai+ai;
      A[1]+= ai;
#else
      A[3]+= ai;
      A[5]+= ai;
      ai   = a*x.A[1];
      A[6]+= ai+ai+ai;
      A[1]+= ai;
      A[8]+= ai;
      ai   = a*x.A[2];
      A[9]+= ai+ai+ai;
      A[2]+= ai;
      A[7]+= ai;
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &add_quadrup(c_evec x, c_real a) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi)                                  
      register real ai = a*x.A[0];
      A[0]+= ai+ai+ai;
#if NDIM==2
      A[2]+= ai;
      ai   = a*x.A[1];
      A[3]+= ai+ai+ai;
      A[1]+= ai;
#else
      A[3]+= ai;
      A[5]+= ai;
      ai   = a*x.A[1];
      A[6]+= ai+ai+ai;
      A[1]+= ai;
      A[8]+= ai;
      ai   = a*x.A[2];
      A[9]+= ai+ai+ai;
      A[2]+= ai;
      A[7]+= ai;
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &quadrup(c_vect x, c_real a, c_real b) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi) + b * Xi*Xj*Xk                   
      register real a3=a+a+a, tmp=b*x.A[0]*x.A[0];
      A[0] = (a3+tmp)*x.A[0];
      A[1] = (a +tmp)*x.A[1];
#if NDIM==2
      tmp  = b*x.A[1]*x.A[1];
      A[2] = (a +tmp)*x.A[0];
      A[3] = (a3+tmp)*x.A[1];
#else
      A[2] = (a +tmp)*x.A[2];
      tmp  = b*x.A[1]*x.A[1];
      A[3] = (a +tmp)*x.A[0];
      A[6] = (a3+tmp)*x.A[1];
      A[7] = (a +tmp)*x.A[2];
      tmp  = b*x.A[2]*x.A[2];
      A[5] = (a +tmp)*x.A[0];
      A[8] = (a +tmp)*x.A[1];
      A[9] = (a3+tmp)*x.A[2];
      A[4] = b*x.A[0]*x.A[1]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    tens &quadrup(c_evec x, c_real a, c_real b) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi) + b * Xi*Xj*Xk                   
      register real a3=a+a+a, tmp=b*x.A[0]*x.A[0];
      A[0] = (a3+tmp)*x.A[0];
      A[1] = (a +tmp)*x.A[1];
#if NDIM==2
      tmp  = b*x.A[1]*x.A[1];
      A[2] = (a +tmp)*x.A[0];
      A[3] = (a3+tmp)*x.A[1];
#else
      A[2] = (a +tmp)*x.A[2];
      tmp  = b*x.A[1]*x.A[1];
      A[3] = (a +tmp)*x.A[0];
      A[6] = (a3+tmp)*x.A[1];
      A[7] = (a +tmp)*x.A[2];
      tmp  = b*x.A[2]*x.A[2];
      A[5] = (a +tmp)*x.A[0];
      A[8] = (a +tmp)*x.A[1];
      A[9] = (a3+tmp)*x.A[2];
      A[4] = b*x.A[0]*x.A[1]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    tens &add_quadrup(c_vect x, c_real a, c_real b) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi) + b * Xi*Xj*Xk                   
      register real a3=a+a+a, tmp=b*x.A[0]*x.A[0];
      A[0]+= (a3+tmp)*x.A[0];
      A[1]+= (a +tmp)*x.A[1];
#if NDIM==2
      tmp  = b*x.A[1]*x.A[1];
      A[2]+= (a +tmp)*x.A[0];
      A[3]+= (a3+tmp)*x.A[1];
#else
      A[2]+= (a +tmp)*x.A[2];
      tmp  = b*x.A[1]*x.A[1];
      A[3]+= (a +tmp)*x.A[0];
      A[6]+= (a3+tmp)*x.A[1];
      A[7]+= (a +tmp)*x.A[2];
      tmp  = b*x.A[2]*x.A[2];
      A[5]+= (a +tmp)*x.A[0];
      A[8]+= (a +tmp)*x.A[1];
      A[9]+= (a3+tmp)*x.A[2];
      A[4]+= b*x.A[0]*x.A[1]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    tens &add_quadrup(c_evec x, c_real a, c_real b) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi) + b * Xi*Xj*Xk                   
      register real a3=a+a+a, tmp=b*x.A[0]*x.A[0];
      A[0]+= (a3+tmp)*x.A[0];
      A[1]+= (a +tmp)*x.A[1];
#if NDIM==2
      tmp  = b*x.A[1]*x.A[1];
      A[2]+= (a +tmp)*x.A[0];
      A[3]+= (a3+tmp)*x.A[1];
#else
      A[2]+= (a +tmp)*x.A[2];
      tmp  = b*x.A[1]*x.A[1];
      A[3]+= (a +tmp)*x.A[0];
      A[6]+= (a3+tmp)*x.A[1];
      A[7]+= (a +tmp)*x.A[2];
      tmp  = b*x.A[2]*x.A[2];
      A[5]+= (a +tmp)*x.A[0];
      A[8]+= (a +tmp)*x.A[1];
      A[9]+= (a3+tmp)*x.A[2];
      A[4]+= b*x.A[0]*x.A[1]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    // const methods                                                            
    //--------------------------------------------------------------------------
    const real* const_pointer()     const { return A; }
    real const&OP[](c_indx i) const { return A[i]; }
    tens OP*       (c_real x) const { register tens y(*this); y*=x; return y; }
    tens OP/       (c_real x) const { register tens y(*this); y*=1/x; return y; }
    tens OP+       (c_tens x) const { register tens y(*this); y+=x; return y; }
    tens OP+       (c_eten x) const { register tens y(*this); y+=x; return y; }
    tens OP-       (c_tens x) const { register tens y(*this); y-=x; return y; }
    tens OP-       (c_eten x) const { register tens y(*this); y-=x; return y; }
    tens OP-       ()         const { register tens y(0); y-=*this; return y; }
    bool OP==      (c_real x) const { return CX(==,x,&&); }
    bool OP==      (c_tens x) const { return CI(==,x.A,&&); }
    bool OP==      (c_eten x) const { return CI(==,x.A,&&); }
    bool OP!=      (c_real x) const { return CX(!=,x,||); }
    bool OP!=      (c_tens x) const { return CI(!=,x.A,||); }
    bool OP!=      (c_eten x) const { return CI(!=,x.A,||); }
    //--------------------------------------------------------------------------
    vect trace() const {
      // return_k = Aijk * D_ij                                                 
      register vect w;
#if NDIM==2
      w.A[0] = A[0]+A[2];
      w.A[1] = A[1]+A[3];
#else
      w.A[0] = A[0]+A[3]+A[5];
      w.A[1] = A[1]+A[6]+A[8];
      w.A[2] = A[3]+A[7]+A[9];
#endif
      return w;
    }
    //--------------------------------------------------------------------------
    real trace(c_vect x) const {
      // trace(x*A) = Dij * xk * Aijk                                           
#if NDIM==2
      return   x.A[0]*(A[0]+A[2]) + x.A[1]*(A[1]+A[3]);
#else
      return   x.A[0]*(A[0]+A[3]+A[5])
	     + x.A[1]*(A[1]+A[6]+A[8])
	     + x.A[2]*(A[2]+A[7]+A[9]);
#endif
    }
    //--------------------------------------------------------------------------
    real trace(c_evec x) const {
      // trace(x*A) = Dij * xk * Aijk                                           
#if NDIM==2
      return   x.A[0]*(A[0]+A[2]) + x.A[1]*(A[1]+A[3]);
#else
      return   x.A[0]*(A[0]+A[3]+A[5])
	     + x.A[1]*(A[1]+A[6]+A[8])
	     + x.A[2]*(A[2]+A[7]+A[9]);
#endif
    }
    //--------------------------------------------------------------------------
    vect OP* (c_matr x) const {
      // return_k = A_ijk*Xij                                                   
      register vect w;
#if NDIM==2
      w.A[0] = A[0]*x.A[0] + A[2]*x.A[2] + twice(A[1]*x.A[1]);
      w.A[1] = A[1]*x.A[0] + A[3]*x.A[2] + twice(A[2]*x.A[1]);
#else
      w.A[0] = A[0]*x.A[0] +A[3]*x.A[3] +A[5]*x.A[5]
	+twice(A[1]*x.A[1] +A[2]*x.A[2] +A[4]*x.A[4]);
      w.A[1] = A[1]*x.A[0] +A[6]*x.A[3] +A[8]*x.A[5]
	+twice(A[3]*x.A[1] +A[4]*x.A[2] +A[7]*x.A[4]);
      w.A[2] = A[2]*x.A[0] +A[7]*x.A[3] +A[9]*x.A[5]
	+twice(A[4]*x.A[1] +A[5]*x.A[2] +A[8]*x.A[4]);
#endif
      return w;
    }
    //--------------------------------------------------------------------------
    vect OP* (c_emat x) const {
      // return_k = A_ijk*Xij                                                   
      register vect w;
#if NDIM==2
      w.A[0] = A[0]*x.A[0] + A[2]*x.A[2] + twice(A[1]*x.A[1]);
      w.A[1] = A[1]*x.A[0] + A[3]*x.A[2] + twice(A[2]*x.A[1]);
#else
      w.A[0] = A[0]*x.A[0] +A[3]*x.A[3] +A[5]*x.A[5]
	+twice(A[1]*x.A[1] +A[2]*x.A[2] +A[4]*x.A[4]);
      w.A[1] = A[1]*x.A[0] +A[6]*x.A[3] +A[8]*x.A[5]
	+twice(A[3]*x.A[1] +A[4]*x.A[2] +A[7]*x.A[4]);
      w.A[2] = A[2]*x.A[0] +A[7]*x.A[3] +A[9]*x.A[5]
	+twice(A[4]*x.A[1] +A[5]*x.A[2] +A[8]*x.A[4]);
#endif
      return w;
    }
    //--------------------------------------------------------------------------
    matr OP* (c_vect x) const {
      // return_ij = A_ijk*Xk                                                   
      register matr w;
#if NDIM==2
      w.A[0] = A[0]*x.A[0] +A[1]*x.A[1];
      w.A[1] = A[1]*x.A[0] +A[2]*x.A[1];
      w.A[2] = A[2]*x.A[0] +A[3]*x.A[1];
#else
      w.A[0] = A[0]*x.A[0] + A[1]*x.A[1] + A[2]*x.A[2];
      w.A[1] = A[1]*x.A[0] + A[3]*x.A[1] + A[4]*x.A[2];
      w.A[2] = A[2]*x.A[0] + A[4]*x.A[1] + A[5]*x.A[2];
      w.A[3] = A[3]*x.A[0] + A[6]*x.A[1] + A[7]*x.A[2];
      w.A[4] = A[4]*x.A[0] + A[7]*x.A[1] + A[8]*x.A[2];
      w.A[5] = A[5]*x.A[0] + A[8]*x.A[1] + A[9]*x.A[2];
#endif
       return w;
    }
    //--------------------------------------------------------------------------
    matr OP* (c_evec x) const {
      // return_ij = A_ijk*Xk                                                   
      register matr w;
#if NDIM==2
      w.A[0] = A[0]*x.A[0] +A[1]*x.A[1];
      w.A[1] = A[1]*x.A[0] +A[2]*x.A[1];
      w.A[2] = A[2]*x.A[0] +A[3]*x.A[1];
#else
      w.A[0] = A[0]*x.A[0] + A[1]*x.A[1] + A[2]*x.A[2];
      w.A[1] = A[1]*x.A[0] + A[3]*x.A[1] + A[4]*x.A[2];
      w.A[2] = A[2]*x.A[0] + A[4]*x.A[1] + A[5]*x.A[2];
      w.A[3] = A[3]*x.A[0] + A[6]*x.A[1] + A[7]*x.A[2];
      w.A[4] = A[4]*x.A[0] + A[7]*x.A[1] + A[8]*x.A[2];
      w.A[5] = A[5]*x.A[0] + A[8]*x.A[1] + A[9]*x.A[2];
#endif
       return w;
    }
    //--------------------------------------------------------------------------
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::esym3                                                          
  //----------------------------------------------------------------------------
  // a symmetric 3x3x3 tensor                                                   
  // element mapping is as follows:   0 -> 0,0,0                                
  //                                  1 -> 0,0,1 & permutations                 
  //                                  2 -> 0,0,2 & permutations                 
  //                                  3 -> 0,1,1 & permutations                 
  //                                  4 -> 0,1,2 & permutations                 
  //                                  5 -> 0,2,2 & permutations                 
  //                                  6 -> 1,1,1                                
  //                                  7 -> 1,1,2 & permutations                 
  //                                  8 -> 1,2,2 & permutations                 
  //                                  9 -> 2,2,2                                
  //----------------------------------------------------------------------------
  // or a symmetric 2x2x2 tensor                                                
  // element mapping is as follows:   0 -> 0,0,0                                
  //                                  1 -> 0,0,1 & permutations                 
  //                                  2 -> 0,1,1 & permutations                 
  //                                  3 -> 1,1,1                                
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL>                      // scalar type                   
  class esym3 {
    //--------------------------------------------------------------------------
    // friendships                                                              
    //--------------------------------------------------------------------------
    friend class sym3<REAL>;
    //--------------------------------------------------------------------------
    // types                                                                    
    //--------------------------------------------------------------------------
  protected:
    typedef REAL                    real;      // type for const returns        
    typedef vector<REAL>            vect;      // type for const returns        
    typedef sym2<REAL>              matr;      // type for const args           
    typedef sym3<REAL>              tens;      // type for const args           
    typedef esym3                   eten;      // type for non-const returns    
    typedef const int             c_indx;      // type for const args           
    typedef const REAL           &c_real;      // type for const args           
    typedef const vector<REAL>   &c_vect;      // type for const args           
    typedef const evector<REAL>  &c_evec;      // type for const args           
    typedef const sym2<REAL>     &c_matr;      // type for const args           
    typedef const esym2<REAL>    &c_emat;      // type for const args           
    typedef const sym3<REAL>     &c_tens;      // type for const args           
    typedef const esym3          &c_eten;      // type for const args           
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
  private:
    REAL *A;
    //--------------------------------------------------------------------------
    // construction                                                             
    //--------------------------------------------------------------------------
  public:
    esym3 (real *x) : A(x) {}
    //--------------------------------------------------------------------------
    // non-const methods                                                        
    //--------------------------------------------------------------------------
  public:
    real* pointer()                   { return A; }
    real &OP[]   (c_indx i)           { return A[i]; }
    eten &OP+=   (c_tens x)           { FI(+=,x.A); return *this; }
    eten &OP+=   (c_eten x)           { FI(+=,x.A); return *this; }
    eten &OP-=   (c_tens x)           { FI(-=,x.A); return *this; }
    eten &OP-=   (c_eten x)           { FI(-=,x.A); return *this; }
    eten &OP*=   (c_real x)           { FX(*=,x); return *this; }
    eten &OP/=   (c_real x)           { return OP*=(real(1)/x); }
    eten &negate ()                   { FI(=-,A); return *this; }
    eten &add_mul(c_tens y, c_real x) { FI(+=,x*y.A); return *this; }
    eten &add_mul(c_eten y, c_real x) { FI(+=,x*y.A); return *this; }
    eten &sub_mul(c_tens y, c_real x) { FI(-=,x*y.A); return *this; }
    eten &sub_mul(c_eten y, c_real x) { FI(-=,x*y.A); return *this; }
    //--------------------------------------------------------------------------
#if NDIM==2
    eten &unity  ()           { A[0]=A[3]=1; A[1]=A[2]=0; return *this; }
    eten &unity  (c_real x)   { A[0]=A[3]=x; A[1]=A[2]=0; return *this; }
    eten &add_unity(c_real x) { A[0]+=x; A[3]+=x;         return *this; }
#else
    //--------------------------------------------------------------------------
    eten &unity  () {
      A[1]=A[2]=A[3]=A[4]=A[5]=A[7]=A[8]=0;
      A[0]=A[6]=A[9]=1;
      return *this;
    }
    eten &unity  (c_real x) {
      A[1]=A[2]=A[3]=A[4]=A[5]=A[7]=A[8]=0;
      A[0]=A[6]=A[9]=x;
      return *this;
    }
    eten &add_unity  (c_real x) {
      A[0]+= x;
      A[6]+= x;
      A[9]+= x;
      return *this;
    }
#endif
    //--------------------------------------------------------------------------
    eten &apply  (real(*f)(c_real))   { 
      A[0]=f(A[0]); A[1]=f(A[1]); A[2]=f(A[2]); A[3]=f(A[3]); 
#if NDIM==3
      A[4]=f(A[4]); A[5]=f(A[5]); A[6]=f(A[6]); A[7]=f(A[7]);
      A[8]=f(A[8]); A[9]=f(A[9]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    eten &connect(c_tens x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]);
      A[1]=f(A[1],x.A[1]);
      A[2]=f(A[2],x.A[2]);
      A[3]=f(A[3],x.A[3]); 
#if NDIM==3
      A[4]=f(A[4],x.A[4]);
      A[6]=f(A[5],x.A[6]);
      A[7]=f(A[7],x.A[7]);
      A[8]=f(A[8],x.A[8]);
      A[9]=f(A[9],x.A[9]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    eten &connect(c_eten x, real(*f)(c_real, c_real)) {
      A[0]=f(A[0],x.A[0]);
      A[1]=f(A[1],x.A[1]);
      A[2]=f(A[2],x.A[2]);
      A[3]=f(A[3],x.A[3]); 
#if NDIM==3
      A[4]=f(A[4],x.A[4]);
      A[6]=f(A[5],x.A[6]);
      A[7]=f(A[7],x.A[7]);
      A[8]=f(A[8],x.A[8]);
      A[9]=f(A[9],x.A[9]);
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    eten &outprod(c_vect x) {
      // outer product Aijk= Xi*Xj*Xk                                           
      register real tmp=x.A[0]*x.A[0];
      A[0] = tmp*x.A[0];
      A[1] = tmp*x.A[1];
#if NDIM==2
      tmp  = x.A[1]*x.A[1];
      A[2] = tmp*x.A[0];
      A[3] = tmp*x.A[1];
#else
      A[2] = tmp*x.A[2];
      tmp  = x.A[1]*x.A[1];
      A[3] = tmp*x.A[0];
      A[4] = x.A[0]*x.A[1]*x.A[2];
      A[5] = tmp*x.A[0];
      A[6] = tmp*x.A[1];
      tmp  = x.A[2]*x.A[2];
      A[7] = tmp*x.A[2];
      A[8] = tmp*x.A[1];
      A[9] = tmp*x.A[2];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    eten &outprod(c_evec x) {
      // outer product Aijk= Xi*Xj*Xk                                           
      register real tmp=x.A[0]*x.A[0];
      A[0] = tmp*x.A[0];
      A[1] = tmp*x.A[1];
#if NDIM==2
      tmp  = x.A[1]*x.A[1];
      A[2] = tmp*x.A[0];
      A[3] = tmp*x.A[1];
#else
      A[2] = tmp*x.A[2];
      tmp  = x.A[1]*x.A[1];
      A[3] = tmp*x.A[0];
      A[4] = x.A[0]*x.A[1]*x.A[2];
      A[5] = tmp*x.A[0];
      A[6] = tmp*x.A[1];
      tmp  = x.A[2]*x.A[2];
      A[7] = tmp*x.A[2];
      A[8] = tmp*x.A[1];
      A[9] = tmp*x.A[2];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &outprod(c_vect x, c_matr y) {
      // outer product Aij = Xi*Yjk + Xj*Yik + Xk*Yij                           
      A[0] = trice(y.A[0]*x.A[0]);
      A[1] = twice(y.A[1]*x.A[0]) + y.A[0]*x.A[1];
#if NDIM==2
      A[2] = twice(y.A[1]*x.A[1]) + y.A[2]*x.A[0];
      A[3] = trice(y.A[2]*x.A[1]);
#else
      A[6] = trice(y.A[3]*x.A[1]);
      A[9] = trice(y.A[5]*x.A[2]);
      A[2] = twice(y.A[2]*x.A[0]) + y.A[0]*x.A[2];
      A[3] = twice(y.A[1]*x.A[1]) + y.A[3]*x.A[0];
      A[5] = twice(y.A[2]*x.A[2]) + y.A[5]*x.A[0];
      A[7] = twice(y.A[4]*x.A[1]) + y.A[3]*x.A[2];
      A[8] = twice(y.A[4]*x.A[2]) + y.A[5]*x.A[1];
      A[4] = y.A[1]*x.A[2] +y.A[2]*x.A[1] +y.A[4]*x.A[0];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &outprod(c_vect x, c_emat y) {
      // outer product Aij = Xi*Yjk + Xj*Yik + Xk*Yij                           
      A[0] = trice(y.A[0]*x.A[0]);
      A[1] = twice(y.A[1]*x.A[0]) + y.A[0]*x.A[1];
#if NDIM==2
      A[2] = twice(y.A[1]*x.A[1]) + y.A[2]*x.A[0];
      A[3] = trice(y.A[2]*x.A[1]);
#else
      A[6] = trice(y.A[3]*x.A[1]);
      A[9] = trice(y.A[5]*x.A[2]);
      A[2] = twice(y.A[2]*x.A[0]) + y.A[0]*x.A[2];
      A[3] = twice(y.A[1]*x.A[1]) + y.A[3]*x.A[0];
      A[5] = twice(y.A[2]*x.A[2]) + y.A[5]*x.A[0];
      A[7] = twice(y.A[4]*x.A[1]) + y.A[3]*x.A[2];
      A[8] = twice(y.A[4]*x.A[2]) + y.A[5]*x.A[1];
      A[4] = y.A[1]*x.A[2] +y.A[2]*x.A[1] +y.A[4]*x.A[0];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &outprod(c_evec x, c_matr y) {
      // outer product Aij = Xi*Yjk + Xj*Yik + Xk*Yij                           
      A[0] = trice(y.A[0]*x.A[0]);
      A[1] = twice(y.A[1]*x.A[0]) + y.A[0]*x.A[1];
#if NDIM==2
      A[2] = twice(y.A[1]*x.A[1]) + y.A[2]*x.A[0];
      A[3] = trice(y.A[2]*x.A[1]);
#else
      A[6] = trice(y.A[3]*x.A[1]);
      A[9] = trice(y.A[5]*x.A[2]);
      A[2] = twice(y.A[2]*x.A[0]) + y.A[0]*x.A[2];
      A[3] = twice(y.A[1]*x.A[1]) + y.A[3]*x.A[0];
      A[5] = twice(y.A[2]*x.A[2]) + y.A[5]*x.A[0];
      A[7] = twice(y.A[4]*x.A[1]) + y.A[3]*x.A[2];
      A[8] = twice(y.A[4]*x.A[2]) + y.A[5]*x.A[1];
      A[4] = y.A[1]*x.A[2] +y.A[2]*x.A[1] +y.A[4]*x.A[0];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &outprod(c_evec x, c_emat y) {
      // outer product Aij = Xi*Yjk + Xj*Yik + Xk*Yij                           
      A[0] = trice(y.A[0]*x.A[0]);
      A[1] = twice(y.A[1]*x.A[0]) + y.A[0]*x.A[1];
#if NDIM==2
      A[2] = twice(y.A[1]*x.A[1]) + y.A[2]*x.A[0];
      A[3] = trice(y.A[2]*x.A[1]);
#else
      A[6] = trice(y.A[3]*x.A[1]);
      A[9] = trice(y.A[5]*x.A[2]);
      A[2] = twice(y.A[2]*x.A[0]) + y.A[0]*x.A[2];
      A[3] = twice(y.A[1]*x.A[1]) + y.A[3]*x.A[0];
      A[5] = twice(y.A[2]*x.A[2]) + y.A[5]*x.A[0];
      A[7] = twice(y.A[4]*x.A[1]) + y.A[3]*x.A[2];
      A[8] = twice(y.A[4]*x.A[2]) + y.A[5]*x.A[1];
      A[4] = y.A[1]*x.A[2] +y.A[2]*x.A[1] +y.A[4]*x.A[0];
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &quadrup(c_vect x, c_real a) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi)                                  
      register real ai = a*x.A[0];
      A[0] = ai+ai+ai;
#if NDIM==2
      A[2] = ai;
      ai   = a*x.A[1];
      A[3] = ai+ai+ai;
      A[1] = ai;
#else
      A[3] = ai;
      A[5] = ai;
      ai   = a*x.A[1];
      A[6] = ai+ai+ai;
      A[1] = ai;
      A[8] = ai;
      ai   = a*x.A[2];
      A[9] = ai+ai+ai;
      A[2] = ai;
      A[4] = 0;
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &quadrup(c_evec x, c_real a) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi)                                  
      register real ai = a*x.A[0];
      A[0] = ai+ai+ai;
#if NDIM==2
      A[2] = ai;
      ai   = a*x.A[1];
      A[3] = ai+ai+ai;
      A[1] = ai;
#else
      A[3] = ai;
      A[5] = ai;
      ai   = a*x.A[1];
      A[6] = ai+ai+ai;
      A[1] = ai;
      A[8] = ai;
      ai   = a*x.A[2];
      A[9] = ai+ai+ai;
      A[2] = ai;
      A[4] = 0;
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &add_quadrup(c_vect x, c_real a) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi)                                  
      register real ai = a*x.A[0];
      A[0]+= ai+ai+ai;
#if NDIM==2
      A[2]+= ai;
      ai   = a*x.A[1];
      A[3]+= ai+ai+ai;
      A[1]+= ai;
#else
      A[3]+= ai;
      A[5]+= ai;
      ai   = a*x.A[1];
      A[6]+= ai+ai+ai;
      A[1]+= ai;
      A[8]+= ai;
      ai   = a*x.A[2];
      A[9]+= ai+ai+ai;
      A[2]+= ai;
      A[7]+= ai;
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &add_quadrup(c_evec x, c_real a) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi)                                  
      register real ai = a*x.A[0];
      A[0]+= ai+ai+ai;
#if NDIM==2
      A[2]+= ai;
      ai   = a*x.A[1];
      A[3]+= ai+ai+ai;
      A[1]+= ai;
#else
      A[3]+= ai;
      A[5]+= ai;
      ai   = a*x.A[1];
      A[6]+= ai+ai+ai;
      A[1]+= ai;
      A[8]+= ai;
      ai   = a*x.A[2];
      A[9]+= ai+ai+ai;
      A[2]+= ai;
      A[7]+= ai;
#endif
      return *this;
    }
    //--------------------------------------------------------------------------
    tens &quadrup(c_vect x, c_real a, c_real b) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi) + b * Xi*Xj*Xk                   
      register real a3=a+a+a, tmp=b*x.A[0]*x.A[0];
      A[0] = (a3+tmp)*x.A[0];
      A[1] = (a +tmp)*x.A[1];
#if NDIM==2
      tmp  = b*x.A[1]*x.A[1];
      A[2] = (a +tmp)*x.A[0];
      A[3] = (a3+tmp)*x.A[1];
#else
      A[2] = (a +tmp)*x.A[2];
      tmp  = b*x.A[1]*x.A[1];
      A[3] = (a +tmp)*x.A[0];
      A[6] = (a3+tmp)*x.A[1];
      A[7] = (a +tmp)*x.A[2];
      tmp  = b*x.A[2]*x.A[2];
      A[5] = (a +tmp)*x.A[0];
      A[8] = (a +tmp)*x.A[1];
      A[9] = (a3+tmp)*x.A[2];
      A[4] = b*x.A[0]*x.A[1]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    tens &quadrup(c_evec x, c_real a, c_real b) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi) + b * Xi*Xj*Xk                   
      register real a3=a+a+a, tmp=b*x.A[0]*x.A[0];
      A[0] = (a3+tmp)*x.A[0];
      A[1] = (a +tmp)*x.A[1];
#if NDIM==2
      tmp  = b*x.A[1]*x.A[1];
      A[2] = (a +tmp)*x.A[0];
      A[3] = (a3+tmp)*x.A[1];
#else
      A[2] = (a +tmp)*x.A[2];
      tmp  = b*x.A[1]*x.A[1];
      A[3] = (a +tmp)*x.A[0];
      A[6] = (a3+tmp)*x.A[1];
      A[7] = (a +tmp)*x.A[2];
      tmp  = b*x.A[2]*x.A[2];
      A[5] = (a +tmp)*x.A[0];
      A[8] = (a +tmp)*x.A[1];
      A[9] = (a3+tmp)*x.A[2];
      A[4] = b*x.A[0]*x.A[1]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    eten &add_quadrup(c_vect x, c_real a, c_real b) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi) + b * Xi*Xj*Xk                   
      register real a3=a+a+a, tmp=b*x.A[0]*x.A[0];
      A[0]+= (a3+tmp)*x.A[0];
      A[1]+= (a +tmp)*x.A[1];
#if NDIM==2
      tmp  = b*x.A[1]*x.A[1];
      A[2]+= (a +tmp)*x.A[0];
      A[3]+= (a3+tmp)*x.A[1];
#else
      A[2]+= (a +tmp)*x.A[2];
      tmp  = b*x.A[1]*x.A[1];
      A[3]+= (a +tmp)*x.A[0];
      A[6]+= (a3+tmp)*x.A[1];
      A[7]+= (a +tmp)*x.A[2];
      tmp  = b*x.A[2]*x.A[2];
      A[5]+= (a +tmp)*x.A[0];
      A[8]+= (a +tmp)*x.A[1];
      A[9]+= (a3+tmp)*x.A[2];
      A[4]+= b*x.A[0]*x.A[1]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    eten &add_quadrup(c_evec x, c_real a, c_real b) {
      // Aijk = a * (Dij*Xk + Dik*Xj + Djk*Xi) + b * Xi*Xj*Xk                   
      register real a3=a+a+a, tmp=b*x.A[0]*x.A[0];
      A[0]+= (a3+tmp)*x.A[0];
      A[1]+= (a +tmp)*x.A[1];
#if NDIM==2
      tmp  = b*x.A[1]*x.A[1];
      A[2]+= (a +tmp)*x.A[0];
      A[3]+= (a3+tmp)*x.A[1];
#else
      A[2]+= (a +tmp)*x.A[2];
      tmp  = b*x.A[1]*x.A[1];
      A[3]+= (a +tmp)*x.A[0];
      A[6]+= (a3+tmp)*x.A[1];
      A[7]+= (a +tmp)*x.A[2];
      tmp  = b*x.A[2]*x.A[2];
      A[5]+= (a +tmp)*x.A[0];
      A[8]+= (a +tmp)*x.A[1];
      A[9]+= (a3+tmp)*x.A[2];
      A[4]+= b*x.A[0]*x.A[1]*x.A[2];
#endif
      return *this;
    }      
    //--------------------------------------------------------------------------
    // const methods                                                            
    //--------------------------------------------------------------------------
    const real* const_pointer()     const { return A; }
    real const&OP[](c_indx i) const { return A[i]; }
    tens OP*       (c_real x) const { register tens y(*this); y*=x; return y; }
    tens OP/       (c_real x) const { register tens y(*this); y*=1/x; return y; }
    tens OP+       (c_tens x) const { register tens y(*this); y+=x; return y; }
    tens OP+       (c_eten x) const { register tens y(*this); y+=x; return y; }
    tens OP-       (c_tens x) const { register tens y(*this); y-=x; return y; }
    tens OP-       (c_eten x) const { register tens y(*this); y-=x; return y; }
    tens OP-       ()         const { register tens y(0); y-=*this; return y; }
    bool OP==      (c_real x) const { return CX(==,x,&&); }
    bool OP==      (c_tens x) const { return CI(==,x.A,&&); }
    bool OP==      (c_eten x) const { return CI(==,x.A,&&); }
    bool OP!=      (c_real x) const { return CX(!=,x,||); }
    bool OP!=      (c_tens x) const { return CI(!=,x.A,||); }
    bool OP!=      (c_eten x) const { return CI(!=,x.A,||); }
    //--------------------------------------------------------------------------
    vect trace() const {
      // return_k = Aijk * D_ij                                                 
      register vect w;
#if NDIM==2
      w.A[0] = A[0]+A[2];
      w.A[1] = A[1]+A[3];
#else
      w.A[0] = A[0]+A[3]+A[5];
      w.A[1] = A[1]+A[6]+A[8];
      w.A[2] = A[3]+A[7]+A[9];
#endif
      return w;
    }
    //--------------------------------------------------------------------------
    real trace(c_vect x) const {
      // trace(x*A) = Dij * xk * Aijk                                           
#if NDIM==2
      return   x.A[0]*(A[0]+A[2]) + x.A[1]*(A[1]+A[3]);
#else
      return   x.A[0]*(A[0]+A[3]+A[5])
	     + x.A[1]*(A[1]+A[6]+A[8])
	     + x.A[2]*(A[2]+A[7]+A[9]);
#endif
    }
    //--------------------------------------------------------------------------
    real trace(c_evec x) const {
      // trace(x*A) = Dij * xk * Aijk                                           
#if NDIM==2
      return   x.A[0]*(A[0]+A[2]) + x.A[1]*(A[1]+A[3]);
#else
      return   x.A[0]*(A[0]+A[3]+A[5])
	     + x.A[1]*(A[1]+A[6]+A[8])
	     + x.A[2]*(A[2]+A[7]+A[9]);
#endif
    }
    //--------------------------------------------------------------------------
    vect OP* (c_matr x) const {
      // return_k = A_ijk*Xij                                                   
      register vect w;
#if NDIM==2
      w.A[0] = A[0]*x.A[0] + A[2]*x.A[2] + twice(A[1]*x.A[1]);
      w.A[1] = A[1]*x.A[0] + A[3]*x.A[2] + twice(A[2]*x.A[1]);
#else
      w.A[0] = A[0]*x.A[0] +A[3]*x.A[3] +A[5]*x.A[5]
	+twice(A[1]*x.A[1] +A[2]*x.A[2] +A[4]*x.A[4]);
      w.A[1] = A[1]*x.A[0] +A[6]*x.A[3] +A[8]*x.A[5]
	+twice(A[3]*x.A[1] +A[4]*x.A[2] +A[7]*x.A[4]);
      w.A[2] = A[2]*x.A[0] +A[7]*x.A[3] +A[9]*x.A[5]
	+twice(A[4]*x.A[1] +A[5]*x.A[2] +A[8]*x.A[4]);
#endif
      return w;
    }
    //--------------------------------------------------------------------------
    vect OP* (c_emat x) const {
      // return_k = A_ijk*Xij                                                   
      register vect w;
#if NDIM==2
      w.A[0] = A[0]*x.A[0] + A[2]*x.A[2] + twice(A[1]*x.A[1]);
      w.A[1] = A[1]*x.A[0] + A[3]*x.A[2] + twice(A[2]*x.A[1]);
#else
      w.A[0] = A[0]*x.A[0] +A[3]*x.A[3] +A[5]*x.A[5]
	+twice(A[1]*x.A[1] +A[2]*x.A[2] +A[4]*x.A[4]);
      w.A[1] = A[1]*x.A[0] +A[6]*x.A[3] +A[8]*x.A[5]
	+twice(A[3]*x.A[1] +A[4]*x.A[2] +A[7]*x.A[4]);
      w.A[2] = A[2]*x.A[0] +A[7]*x.A[3] +A[9]*x.A[5]
	+twice(A[4]*x.A[1] +A[5]*x.A[2] +A[8]*x.A[4]);
#endif
      return w;
    }
    //--------------------------------------------------------------------------
    matr OP* (c_vect x) const {
      // return_ij = A_ijk*Xk                                                   
      register matr w;
#if NDIM==2
      w.A[0] = A[0]*x.A[0] +A[1]*x.A[1];
      w.A[1] = A[1]*x.A[0] +A[2]*x.A[1];
      w.A[2] = A[2]*x.A[0] +A[3]*x.A[1];
#else
      w.A[0] = A[0]*x.A[0] + A[1]*x.A[1] + A[2]*x.A[2];
      w.A[1] = A[1]*x.A[0] + A[3]*x.A[1] + A[4]*x.A[2];
      w.A[2] = A[2]*x.A[0] + A[4]*x.A[1] + A[5]*x.A[2];
      w.A[3] = A[3]*x.A[0] + A[6]*x.A[1] + A[7]*x.A[2];
      w.A[4] = A[4]*x.A[0] + A[7]*x.A[1] + A[8]*x.A[2];
      w.A[5] = A[5]*x.A[0] + A[8]*x.A[1] + A[9]*x.A[2];
#endif
       return w;
    }
    //--------------------------------------------------------------------------
    matr OP* (c_evec x) const {
      // return_ij = A_ijk*Xk                                                   
      register matr w;
#if NDIM==2
      w.A[0] = A[0]*x.A[0] +A[1]*x.A[1];
      w.A[1] = A[1]*x.A[0] +A[2]*x.A[1];
      w.A[2] = A[2]*x.A[0] +A[3]*x.A[1];
#else
      w.A[0] = A[0]*x.A[0] + A[1]*x.A[1] + A[2]*x.A[2];
      w.A[1] = A[1]*x.A[0] + A[3]*x.A[1] + A[4]*x.A[2];
      w.A[2] = A[2]*x.A[0] + A[4]*x.A[1] + A[5]*x.A[2];
      w.A[3] = A[3]*x.A[0] + A[6]*x.A[1] + A[7]*x.A[2];
      w.A[4] = A[4]*x.A[0] + A[7]*x.A[1] + A[8]*x.A[2];
      w.A[5] = A[5]*x.A[0] + A[8]*x.A[1] + A[9]*x.A[2];
#endif
      return w;
    }
    //--------------------------------------------------------------------------
  };
  //////////////////////////////////////////////////////////////////////////////
  // related functions (coding as friends results in compiler problems)         
  //////////////////////////////////////////////////////////////////////////////
  template<typename REAL> inline
  vector<REAL> OP*(const sym2<REAL>&x, const sym3<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  vector<REAL> OP*(const sym2<REAL>&x, const esym3<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  vector<REAL> OP*(const esym2<REAL>&x, const sym3<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  vector<REAL> OP*(const esym2<REAL>&x, const esym3<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  sym2<REAL>   OP*(const vector<REAL>&x, const sym3<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  sym2<REAL>   OP*(const vector<REAL>&x, const esym3<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  sym2<REAL>   OP*(const evector<REAL>&x, const sym3<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  sym2<REAL>   OP*(const evector<REAL>&x, const esym3<REAL>&y) { return y*x; }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  std::ostream& OP<<(std::ostream&s, const sym3<REAL>& x) {
    return util::write_array<REAL,NTEN>(s,x.const_pointer());
  }
  //----------------------------------------------------------------------------
  template<typename REAL> inline
  std::ostream& OP<<(std::ostream&s, const esym3<REAL>& x) {
    return util::write_array<REAL,NTEN>(s,x.const_pointer());
  }
  //----------------------------------------------------------------------------
  template<typename A, typename B, typename REAL> inline
  std::istream& OP>>(std::istream&s, sym3<REAL>& x) {
    return util::read_array<REAL,NTEN>(s,x.pointer());
  }
  //----------------------------------------------------------------------------
  template<typename A, typename B, typename REAL> inline
  std::istream& OP>>(std::istream&s, esym3<REAL>& x) {
    return util::read_array<REAL,NTEN>(s,x.pointer());
  }
}
#undef NMAT
#undef CI
#undef CX
#undef FI
#undef FX
#undef OP
#undef NTEN
////////////////////////////////////////////////////////////////////////////////
#endif // included_tens_h
