// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// yanc.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fr Astronomie,                               |
//           Knigstuhl 17, D-69117 Heidelberg, Germany                        |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                        YET ANOTHER N-BODY CODE                              |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class yanc                         see yanc.cc                              |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_yanc_h
#define included_yanc_h

#ifndef included_iostream
#include <iostream>
#define included_iostream
#endif
using std::ostream;

#ifndef included_addo_h
#include <nbdy/addo.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {                               // yanc is in namespace nbdy     
  class ndata;                                 // forward declare ndata         
  class basic_nbody;                           // forward declare nbody         
  class potential;                             // forward declare potential     
  class yanc {
  private:
    ndata       *MYNDATA;                      // our N-body data               
    basic_nbody *MYNBODY;                      // our N-body code               
    yanc();                                    // not implemented               
    yanc(const yanc&);                         // not implemented               
    yanc& operator= (const yanc&);             // not implemented               
  public:
    //--------------------------------------------------------------------------
    yanc(                                      // constructor without parameters
	 const char *);                        // I:   yanc-formatted input file
#ifdef ALLOW_NEMO
    //--------------------------------------------------------------------------
    yanc(                                      // constructor taking parameters 
	 const char*,                          // I:   input file (yanc/nemo)   
	 const bool,                           // I:   do nemo I/O?             
	 const double,                         // I:   theta                    
	 const int,                            // I:   Nreuse                   
	 const int,                            // I:   Ncrit                    
	 const double,                         // I:   eps                      
	 const int,                            // I:   kern                     
	 const int,                            // I:   hmin                     
	 const add_out,                        // I:   what to write out        
	 const int,                            // I:   Nsteps                   
	 const int,                            // I:   Nref                     
	 const double,                         // I:   fac                      
	 const bool       =false,              //[I:   resume old (if nemo I/O) 
	 const potential* =0);                 //[I:   external potential]      
    //--------------------------------------------------------------------------
    void  describe_nemo   (                    // describe simulation           
			   ostream&,           // I:  output stream             
			   const char*);       // I:   command line             
    //--------------------------------------------------------------------------
    void  open_nemo       (                    // open NEMO output stream       
			   const char* =0,     //[I:   file name ]              
			   const bool = false);//[I"   resume old sim?]         
    //--------------------------------------------------------------------------
    void  close_nemo      ();                  // close NEMO output stream      
    void  write_nemo      ();                  // NEMO-format output            
    bool  nemo_is_open    () const;            // ready for NEMO output ?       
#endif // ALLOW_NEMO
    //--------------------------------------------------------------------------
    ~yanc();                                   // destructor                    
    //--------------------------------------------------------------------------
    void  full_step       ();                  // perform one full time step    
    //--------------------------------------------------------------------------
    void  describe        (                    // describe simulation           
			   ostream&,           // I:   output stream            
			   const char* =0,     // [I]: name of main             
			   const char* =0);    // [I]: base name: output files  
    //--------------------------------------------------------------------------
    static
    void  describe_yanc_format(                // describe yanc I/O format      
			       ostream&);      // I/O: stream to write to       
    //--------------------------------------------------------------------------
    bool  write_yanc_ascii(                    // YANC-format output to file    
			   const char*,        // I:   file name                
			   const char* =0);    // [I]:  name of main            
    //--------------------------------------------------------------------------
    bool  write_yanc_binary(                   // write yanc form, bodies binary
			    const char*,       // I:   name of output file      
			    const char* =0)const;// I: name of calling program  
    //--------------------------------------------------------------------------
    void  long_stats      (ostream&) const;    // write long stats              
    void  short_stats     (ostream&) const;    // write short stats             
    void  short_head      (ostream&) const;    // write header for short stats  
    bool  okay            ()         const;    // is all fine?                  
    int   Nsteps          ()         const;    // # time steps levels           
    int   Nbodies         ()         const;    // # bodies                      
    const char* data_file ()         const;    // input data file               
    float eps             ()         const;    // softening length              
    float tau_min         ()         const;    // minimum time step             
    float tau_max         ()         const;    // maximum time step             
    int   softening       ()         const;    // type of softening             
    int   kernel          ()         const;    // type of kernel                
    float time            ()         const;    // current simulation time       
    float initial_time    ()         const;    // time at input                 
    float theta           ()         const;    // opening parameter             
    int   Ncrit           ()         const;    // # max bodies / un-split cell  
    int   Nreuse          ()         const;    // # max subsequent tree re-uses 
    int   Nref            ()         const;    // # bodies in cell for rho estim
    float fac             ()         const;    // time-stepping parameter       
    float cpu_total       ()         const;    // accumulated cpu time [sec]    
    //--------------------------------------------------------------------------
  };
}
//------------------------------------------------------------------------------
#endif // included_yanc_h
